"""
# ============================================================================
# The controller of all the views
#
# Events triggered by ViewsControllerTk
#   call_view(resource)
#
# Events triggered by ViewDynamicTk
#   undo()
#   redo()
#
# Feedback from Views
#   open_companion_view(block)
#
#   notify_block_movable_toggled( block)
#   notify_move_block_to_resource(resource_type,block,from_resource,resource)
#   notify_block_move(self.resource.number, block, gui_block_day, gui_block_start_time)
#
#   save_action_block_movable_toggled(block, not block.movable)
#   save_action_block_move(block, from_day, to_day, from_start, to_start)
#   save_action_block_resource_changed(resource_type, block,from_resource,to_resource)
#
#   remove_all_redoes()
#   view_is_closing(self.resource)
# ============================================================================
"""

from typing import Callable, Optional, Literal

from ..gui_pages.views_controller_tk import ViewsControllerTk
from ..model import ResourceType, Schedule, Stream, Teacher, Lab, Block
from .view import View

RESOURCE = Teacher | Lab | Stream

# =====================================================================================================================
# Actions
# =====================================================================================================================

class Action:
    def __init__(self, block: Block, action: Literal['move', 'change_resource', 'toggle_movable'],
            from_day: Optional[float] = None,
            from_time: Optional[float] = None,
            to_day: Optional[float] = None,
            to_time: Optional[float] = None,
            from_resource: Optional[RESOURCE] = None,
            to_resource: Optional[RESOURCE] = None,
            resource_type:Optional[ResourceType] = None,
            was_movable: bool = None,
                 ):
        self.action = action
        self.block = block
        self.to_day = to_day
        self.to_time = to_time
        self.from_day = from_day
        self.from_time = from_time
        self.from_resource = from_resource
        self.to_resource = to_resource
        self.resource_type = resource_type
        self.was_movable = was_movable

    def __str__(self):
        if self.action == 'move':
            return f"move {self.block} from {self.from_day} to {self.to_day} and from {self.from_time} to {self.to_time}"
        elif self.action == "change_resource":
            return f"{self.block} change from {self.from_resource} to {self.to_resource}"
        elif self.action == "toggle_movable":
            return f"{self.block} movable has been toggled"
        else:
            return "Action with no action"

# =====================================================================================================================
# Views Controller
# =====================================================================================================================
class ViewsController:
    """Controls all of the views and their inter-view communication"""

    # -----------------------------------------------------------------------------------------------------------------
    # init
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, dirty_flag_method: Callable[[Optional[bool]], bool], frame, schedule: Schedule,
                 gui:ViewsControllerTk = None):
        """
        :param dirty_flag_method: method used to indicate that data has changed
        :param frame: where the view choices will be displayed
        :param schedule:
        """
        self.dirty_flag_method=dirty_flag_method
        self.frame = frame
        self.schedule = schedule
        self._views: dict[str, View] = {}
        self._undo: list[Action] = []
        self._redo: list[Action] = []


        self.resources = {
            ResourceType.teacher: list(self.schedule.teachers()),
            ResourceType.lab: list(self.schedule.labs()),
            ResourceType.stream: list(self.schedule.streams())
        }

        if gui is not None:
            self.gui = gui
        else:
            self.gui = ViewsControllerTk(frame, self.resources, self.call_view)


    # -----------------------------------------------------------------------------------------------------------------
    # refresh
    # -----------------------------------------------------------------------------------------------------------------
    def refresh(self):
        """
        sets the button colours for view choices depending on the most severe conflict for that resource
        """

        # if resources have changed, then we need to close all the views, and create a new gui
        resources = {
            ResourceType.teacher: list(self.schedule.teachers()),
            ResourceType.lab: list(self.schedule.labs()),
            ResourceType.stream: list(self.schedule.streams())
        }
        redrawing = False
        for resource_type in ResourceType.teacher, ResourceType.lab, ResourceType.stream:
            if (len(set(self.resources[resource_type]) - set(resources[resource_type])) != 0
                or len(set(resources[resource_type]) - set(self.resources[resource_type])) != 0):
                redrawing = True
                break
        if redrawing:
            self.resources = resources
            self.gui = ViewsControllerTk(self.frame, self.resources, self.call_view)

        # update the colours
        self.schedule.calculate_conflicts()
        for resource_type in self.resources:
            for resource in self.resources[resource_type]:
                conflict = self.schedule.resource_conflict(resource)
                conflict = conflict.most_severe(resource_type)
                self.gui.set_button_colour(resource.number, resource_type, conflict)


    # -----------------------------------------------------------------------------------------------------------------
    # call view
    # -----------------------------------------------------------------------------------------------------------------
    def call_view(self, resource):
        """
        creates the view for this resource, or if it already exists, makes it visible
        :param resource:
        """
        if resource.number in self._views:
            self._views[resource.number].gui.raise_to_top()
            return

        self._views[resource.number] = View(self, self.frame, self.schedule, resource, self.dirty_flag_method)


    # -----------------------------------------------------------------------------------------------------------------
    # notify block move
    # -----------------------------------------------------------------------------------------------------------------
    def notify_block_move(self, resource_number: Optional[str], moved_block: Block, day: float, start_time:float):
        """A GUI block has been modified in a view, so propagate this information to the other views"""

        self.refresh()

        # update any view that has the same block that was moved
        for view_id, view in self._views.items():
            if resource_number is None or view_id != resource_number:
                if view.is_block_in_view(moved_block):
                    view.move_gui_block_to(moved_block, day, start_time)
            view.refresh_block_colours()

    # -----------------------------------------------------------------------------------------------------------------
    # notify block movable toggled
    # -----------------------------------------------------------------------------------------------------------------
    def notify_block_movable_toggled(self, modified_block: Block):
        """A block has been modified in a view, so propagate this information to the other views"""

        self.refresh()

        # update any view that has the same block that was moved
        for view_id, view in self._views.items():
            view.refresh_block_colours()

    # ----------------------------------------------------------------------------------------------------------------
    # notify move block to a different resource
    # ----------------------------------------------------------------------------------------------------------------
    def notify_move_block_to_resource(self, resource_type: ResourceType, block: Block, from_resource: RESOURCE, to_resource: RESOURCE):
        """
        a block has been moved from one resource to another
        :param resource_type:
        :param block:
        :param from_resource:
        :param to_resource:
        """
        match resource_type:
            case ResourceType.teacher:
                block.remove_teacher(from_resource)
                block.add_teacher(to_resource)
            case ResourceType.lab:
                block.remove_lab(from_resource)
                block.add_lab(to_resource)
        self.schedule.calculate_conflicts()

        if from_resource.number in self._views.keys():
            self._views[from_resource.number].draw()
        else:
            self.call_view(from_resource)

        if to_resource.number in self._views.keys():
            self._views[to_resource.number].draw()
        else:
            self.call_view(to_resource)

        self.refresh()

        self.dirty_flag_method(True)


    # ----------------------------------------------------------------------------------------------------------------
    # open companion view (double_click_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def open_companion_view(self, block: Block):
        """
        Based on the resource_type open other views which have this Block.
        :param block:
        """
        teachers = block.teachers()
        labs = block.labs()
        streams = block.streams()

        for resource in (*teachers, *labs, *streams):
            self.call_view(resource)

    # ----------------------------------------------------------------------------------------------------------------
    # view is closing
    # ----------------------------------------------------------------------------------------------------------------
    def view_is_closing(self, resource: RESOURCE):
        """view for resource has been closed, update our internal database"""
        try:
            self._views.pop(resource.number)
        except KeyError:
            pass

    # -----------------------------------------------------------------------------------------------------------------
    # save changes to undo db
    # -----------------------------------------------------------------------------------------------------------------
    def save_action_block_move(self, block: Block, from_day:float, to_day:float, from_time:float, to_time:float):

        # add to undo list
        self._undo.append(Action(block=block, action="move", from_day=from_day, from_time=from_time,
                                 to_day = to_day, to_time = to_time))

    def save_action_block_resource_changed(self, resource_type: ResourceType, block: Block, from_resource, to_resource):

        # add to undo list
        self._undo.append(Action(block=block, action="change_resource",
                                 from_resource=from_resource,
                                 to_resource=to_resource,
                                 resource_type=resource_type))

    def save_action_block_movable_toggled(self, block: Block, was: bool):

        # add to undo list
        self._undo.append(Action(block=block, action="toggle_movable", was_movable=was))


    # ----------------------------------------------------------------------------------------------------------------
    # undo/redo last actions
    # ----------------------------------------------------------------------------------------------------------------
    def undo(self):
        """
        undo whatever action remains in the undo list
        """
        if len(self._undo) == 0:
            return

        action = self._undo.pop()
        self._process_action(action, self._redo)
        self.dirty_flag_method(True)

    def redo(self):
        """
        redo whatever action remains in the redo list
        """
        if len(self._redo) == 0:
            return

        action = self._redo.pop()
        self._process_action(action, self._undo)
        self.dirty_flag_method(True)

    # ----------------------------------------------------------------------------------------------------------------
    # process the undo/redo action
    # ----------------------------------------------------------------------------------------------------------------
    def _process_action(self, action, other_list):
        """
        process the undo/redo action
        :param action: the action to do
        :param other_list: what list to move this action to... undo->redo and vice versa
        :return:
        """
        match action.action:
            case 'move':
                other_list.append(Action(action='move', block= action.block,
                                         from_day = action.to_day, from_time=action.to_time,
                                         to_day = action.from_day, to_time=action.from_time,
                                         ))
                action.block.start = action.from_time
                action.block.snap_to_day(action.from_day)
                self.schedule.calculate_conflicts()
                self.notify_block_move(None, action.block, action.from_day, action.from_time)

            case 'change_resource':
                other_list.append(Action(action='change_resource', block=action.block,
                                         from_resource=action.to_resource,
                                         to_resource=action.from_resource,
                                         resource_type=action.resource_type,
                                  ))
                self.schedule.calculate_conflicts()
                self.notify_move_block_to_resource(resource_type=action.resource_type, block=action.block,
                                                    from_resource=action.to_resource,
                                                    to_resource=action.from_resource)
            case 'toggle_movable':
                other_list.append(Action(action='toggle_movable', block=action.block, was_movable=not action.was_movable))
                action.block.movable = action.was_movable
                self.notify_block_movable_toggled(action.block)



    # ----------------------------------------------------------------------------------------------------------------
    # no more redoes
    # ----------------------------------------------------------------------------------------------------------------
    def remove_all_redoes(self):
        """Remove all redoes from the redo list"""
        self._redo.clear()

    # ----------------------------------------------------------------------------------------------------------------
    # redraw all the views and update button choices
    # ----------------------------------------------------------------------------------------------------------------
    def redraw_all(self):
        """redraw all the views and update button choices"""
        self.schedule.calculate_conflicts()

        for resource, view in self._views.items():
            teacher_numbers = (x.number for x in self.schedule.teachers())
            lab_numbers = (x.number for x in self.schedule.labs())
            stream_numbers = (x.number for x in self.schedule.streams())

            # if the resource for this view has been deleted from the schedule, then close the view
            if resource not in teacher_numbers and resource not in stream_numbers and resource not in lab_numbers:
                view.close()
            else:
                view.draw()

    # ----------------------------------------------------------------------------------------------------------------
    # kill bill!
    # ----------------------------------------------------------------------------------------------------------------
    def kill_all_views(self):
        """vicious!"""
        for view in self._views.values():
            view.close()



from ..Utilities import Colour
from ..model import ResourceType, ConflictType

# ==============================================================================================================
# some colours
# ==============================================================================================================
LIME_GREEN = "#ccffcc"
SKY_BLUE = "#b3e6ff"
TEAL = Colour.add(SKY_BLUE, LIME_GREEN)
IMMOVABLE_COLOUR: str = "#dddddd"

RESOURCE_COLOURS: dict[ResourceType, str] = {
    ResourceType.lab: "#cdefab",
    ResourceType.teacher: "#abcdef",
    ResourceType.stream: TEAL
}

# ==============================================================================================================
# return a list of colours with names, defining what colours go with what conflict types
# ==============================================================================================================
def get_conflict_colour_info(resource_type:ResourceType) -> list [dict[str, str]]:
    """
    :param resource_type: teacher/lab/stream
    :return: a list of dictionaries: bg = background colour, fg = foreground colour, text = conflict type
    """

    conflict_colour_info = []
    conflict_types = [ConflictType.TIME, ConflictType.LUNCH,
              ConflictType.MINIMUM_DAYS, ConflictType.AVAILABILITY]

    match resource_type:
        case ResourceType.teacher:
            conflict_types.insert(0, ConflictType.TIME_TEACHER)
        case ResourceType.stream:
            conflict_types.insert(0, ConflictType.TIME_STREAM)
        case ResourceType.lab:
            conflict_types.insert(0, ConflictType.TIME_LAB)

    for c in conflict_types:
        bg = ConflictType.colours()[c]
        fg = "white"
        if Colour.is_light(bg):
            fg = "black"
        text = c.name
        conflict_colour_info.append({
            "bg": Colour.string(bg),
            "fg": Colour.string(fg),
            "text": text
        })
    return conflict_colour_info




from __future__ import annotations
from typing import *
from tkinter import *
from tkinter import ttk
from functools import partial


class DynamicTree(ttk.Treeview):

    def __init__(self, master, *_, **kwargs):
        super().__init__(master)
        self.master = master
        self._configuration_memory: dict[str, Any] = dict()

        # ---------------------------------------------------------------
        # configuration and defaults
        # IMPORTANT NOTE:  Every key in __config_specs needs a
        #                  corresponding key in _defaults, or the
        #                  program will crash
        # ---------------------------------------------------------------
        # this is a table of all the additional options available to
        # Dynamic course_ttkTreeView and the methods used to set_default_fonts_and_colours those options

        _status_variable: StringVar = StringVar()
        self.__config_specs: dict[str, Callable[[DynamicTree, Any], None]] = {
            'children': partial(self._save_config_spec, 'children'),
            'on_select': partial(self._save_config_spec, 'on_select'),
            'status_var': partial(self._save_config_spec, 'status_var'),
            'right_click': partial(self._save_config_spec, 'status_var'),
        }

        # this is a table of all the additional options available to
        # Dynamic course_ttkTreeView, and their default values
        self._defaults: dict[str, Callable[[DynamicTree, Any], None]]
        self._defaults = {
            'children': lambda children, selected: None,
            'on_select': lambda selected_obj, selected: None,
            'status_var': _status_variable,
            'right_click': lambda selected_obj, selected: None,
        }

        # set options to defaults
        self._configuration_memory = self._defaults.copy()

        # ---------------------------------------------------------------
        # apply user options
        # ---------------------------------------------------------------
        self.configure(**kwargs)

        # ---------------------------------------------------------------
        # setup bindings
        # ---------------------------------------------------------------
        self.bind('<<TreeviewOpen>>', self.open_branch)

    # =================================================================================================
    # configure, cget, etc
    # =================================================================================================
    def _save_config_spec(self, this_property: str, value: Any):
        self._configuration_memory[this_property] = value

    def cget(self, option: str) -> Any:
        if option in self._configuration_memory:
            return self._configuration_memory[option]
        else:
            return super().cget(option)

    def configure(self, **kwargs):
        for k, v in kwargs.items():
            if k in self._configuration_memory:
                self._configuration_memory[k](v)
            else:
                super().configure(**{k: v})

    # ====================================================================================
    # run a callback routine
    # ====================================================================================
    def __run_callback(self, which_one, *args):
        callback: callable = self.cget(which_one)
        if callback:
            callback(*args)

    # =================================================================================================
    # sub open_branch -> callback for course_ttkTreeView
    # -> dynamically updates the next level of grandchildren, using
    #    callback routine defined by '-children'
    # =================================================================================================
    def open_branch(self, *_):
        item = self.focus()


'''

# ----------------------------------------------------------------------
# sub openbranch -> callback for course_ttkTreeView
# -> dynamically updates the next level of grandchildren, using
#    callback routine defined by '-children'
# ----------------------------------------------------------------------
sub OpenCmd {
    
    # only continue on if callback is defined
    return unless $self->cget(-children);

    # get a list of children of this branch (should this be leaf instead of child?)
    my @children = $self->infoChildren($path);

    # loop through all the children
    my $i = 0;
    my $sep = $self->cget(-separator);
    foreach my $child (@children) {
        my @grandkids;
        

        # update status variable
        (my $node) = $child =~ /^.*?([^\Q$sep\E]*)$/;
        if (length($child) < 30) {
            $$status_ptr = "Scanning node $child";
        }
        else {
            $$status_ptr = "Scanning node ... ".substr($child,-30);
        }
        $i++;
        $$percent_ptr = ($i/@children)*100;
        $mw->update() if $mw;

        # have grandchildren already been defined?
        next if $self->infoChildren($child);

        # define grandchildren for this child
        ($node) = $child =~ /([^\Q$sep\E]*)$/;
        @grandkids = $self->Callback(-children=>$self,$child);

        # avoid infinite loop
        if ($child =~ /\Q$sep$node$sep\E/) {
            @grandkids = (['<etc>',{-state=>'disabled'}]);
        }


        foreach my $kid (@grandkids) {
            my $newkid;
            my %kidhash;
            $newkid = $kid;
            if (ref $kid eq 'ARRAY') {
                $newkid = $kid->[0];
                if (ref $kid->[1] eq 'HASH') {
                    %kidhash = %{$kid->[1]};
                }
            }
            $self->add ("$child$sep$newkid",-text=>$newkid, %kidhash)
                    unless $self->infoExists("$child$sep$newkid");
            $self->hide('entry', "$child$sep$newkid");
        }


    }

    # done, update status variable and progress bar
    $self->autosetmode;
    $$status_ptr = "Done";
    $$percent_ptr = 0;
}



sub selectionSet {
    my $self = shift;
    my $input = shift;
    return unless $self->infoExists($input);
    $self->SUPER::selectionSet($input);
    $self->Callback(-onSelect=>$self,$input);
}

# ======================================================================
# Expand Tree given a specific node to time_start from
# ======================================================================
{ my $goforit;
  my $treetext;
    sub ExpandTree {
        my $self     = shift;
        my $maxdepth = shift;
        my $keeptext = shift || 0;
        $treetext = '';

        my $selection = $self->selectionGet;
        my $select; $select = (ref $selection) ? $selection->[0] : $selection;
        return unless $select;

        $goforit = 1;
        return recursive_expand($self,$select,$maxdepth,$keeptext);
    }

    sub TreeText {
        return ExpandTree(@_,1);
    }

    sub recursive_expand {
        my $self = shift;
        my $path = shift;
        my $maxdepth = shift || 0;
        my $keeptext = shift || 0;
        my $curdepth = shift || 0;

        if ($self->getmode($path) ne 'none') {
            $self->open($path);
        }

        # bail out if curdepth exceeds max
        return if $maxdepth && $curdepth > $maxdepth;

        # keep text
        if ($keeptext) {
            my $sep = $self->cget(-separator);
            (my $ele = $path) =~ s/.*?([^$sep]*)$/$1/;
            $treetext = $treetext . "\n" . "  "x$curdepth . $ele;
        }

        # procreate
        $curdepth++;
        foreach my $child ($self->infoChildren($path)) {
            recursive_expand ($self,$child,
                              $maxdepth,$keeptext,$curdepth) if $goforit;
        }
        return $treetext;
    }

    sub CancelExpandTree {
        my $self = shift;
        $goforit = 0;
    }
}

# ======================================================================
# Collapse Tree given a specific node to time_start from
# ======================================================================
sub CollapseTree {
    my $self = shift;
    my $selection = $self->selectionGet;
    my $select; $select = (ref $selection) ? $selection->[0] : $selection;
    return unless $select;
    recursive_collapse($self,$select);
}
sub recursive_collapse {
    my $self = shift;
    my $path = shift;
    $self->close($path);
    foreach my $child($self->infoChildren($path)) {
        recursive_collapse($self,$child);
    }
}
1;
'''



import tkinter
from functools import partial
from tkinter import *
from tkinter.simpledialog import Dialog

print("Testing simple dialog boxes")
mw = Tk()
mw.geometry('500x500')

class A(Dialog):
    # creates parent window
    def body(self, frame:Frame):
        """
        setup body, return widget that will have the initial focus
        :param frame:
        :return:
        """
        self.initialvalue = 3
        w = Label(frame, text="Enter Number", justify="left")
        w.grid(row=0, padx=5, sticky='w')

        self.entry = Entry(frame, name="entry")
        self.entry.grid(row=1, padx=5, sticky="we")

        if self.initialvalue is not None:
            self.entry.insert(0, str(self.initialvalue))
            self.entry.select_range(0, "end")

        return self.entry

    def validate(self):
        try:
            n = self.entry.getint(self.entry.get())
        except ValueError:
            self.entry.delete(0, "end")
            self.entry.insert(0, str(self.initialvalue))
            self.entry.select_range(0, "end")
            return False
        if n == 5:
            return False
        return True

    def ok(self, event=None):
        print("Called ok")
        super().ok(event)

    def apply(self):
        print("Called apply")
        super().apply()

    def cancel(self, event=None):
        print("Called cancel")
        super().cancel(event)

def show_dialog(*_):
    a=A(mw, "This Title")

def is_ok(number,*_):
    if number == "" or number == ".":
        return True
    try:
        float(number)
        return True
    except ValueError:
        duration_entry.winfo_toplevel().bell()
        return False


is_ok_tk = mw.register(is_ok)

Button(mw, text="press me for dialog box", command=show_dialog ).pack()
tk_duration=tkinter.StringVar()
duration_entry = Entry(mw, textvariable=tk_duration, validate='key', validatecommand=(is_ok_tk, '%P', '%s'))
duration_entry.pack()

def print_contents( event):
    print("Hi. The current entry content is:",
          tk_duration.get())

duration_entry.bind('<Key-Return>',
                             print_contents)


mainloop()


from __future__ import annotations

from typing import Any, Optional, TYPE_CHECKING, Iterable

import csv

from . import TimeSlot
from .enums import ResourceType, SemesterType
from .enums import WeekDay

# stuff that we need just for type checking, not for actual functionality

if TYPE_CHECKING:
    from .block import Block
    from .stream import Stream
    from .section import Section
    from .teacher import Teacher
    from .lab import Lab
    from .course import Course
    from .schedule import Schedule


# TODO: not saving or reading synced blocks

class CSVSerializor:
    last_line_number_read:int  = -1
    last_line_read:str = ""

    # ============================================================================
    # write csv file
    # ============================================================================
    @staticmethod
    def write(schedule: Schedule, file: str):
        """Write all the details of the schedule to a file, throws an exception if the file cannot be written to"""

        with open(file, 'w', newline='') as f:
            w = csv.writer(f, delimiter=',',
                           quotechar='|', quoting=csv.QUOTE_MINIMAL)

            # --------------------------------------------------------------------
            # write all the 'collectables' first
            # --------------------------------------------------------------------
            w.writerow([None, 'number', 'description'])
            for lab in sorted(schedule.labs()):
                w.writerow(["lab", lab.number, lab.description])
                for unavail in lab.unavailable_slots():
                    w.writerow(["unavailable", "", unavail.day, unavail.start, unavail.duration,
                                int(unavail.movable)])
            w.writerow([])

            w.writerow([None, 'number', 'description'])
            for stream in sorted(schedule.streams()):
                w.writerow(["stream", stream.number, stream.description])
            w.writerow([])

            w.writerow([None, 'number', 'first name', 'last name', 'department', 'release'])
            for teacher in sorted(schedule.teachers(), key=lambda x: x.number):
                w.writerow(
                    ["teacher", teacher.number, teacher.firstname, teacher.lastname,
                     teacher.department, teacher.release])
            w.writerow([])

            # --------------------------------------------------------------------
            # courses/sections/blocks
            # --------------------------------------------------------------------
            w.writerow([])
            w.writerow([None, None, 'DESCRIPTION', 'OF', 'FIELDS'])
            w.writerow([None, 'number', 'name', 'semester', 'needs allocation', 'hours_per_week', 'COURSE'])
            w.writerow([None, 'id', 'number', 'name', 'students', 'SECTION'])
            w.writerow([None, 'id', 'day', 'start', 'duration', 'movable', 'BLOCK'])
            w.writerow([])
            for course in sorted(schedule.courses()):
                w.writerow(["course", course.number, course.name, course.semester.value,
                            int(course.needs_allocation), course.hours_per_week])

                for section in course.sections():
                    w.writerow([])
                    w.writerow(["section", None, section.number, section.name, section.num_students])

                    for s in section.streams():
                        w.writerow(["add_stream", s.number])

                    for block in section.blocks():
                        w.writerow(["add_block", None, block.day.name, block.start,
                                    block.duration, int(block.movable)])
                        for teacher in sorted(block.teachers(), key=lambda t: t.number):
                            w.writerow(["add_block_teacher", teacher.number])
                        for lab in sorted(block.labs(), key=lambda ll: ll.number):
                            w.writerow(["add_lab", lab.number])

                    if len(section.section_defined_teachers()) != 0:
                        w.writerow([])
                        w.writerow([None, None, None, 'assigne allocation, but not assigned to any blocks'])
                        w.writerow([None, 'id', 'allocation'])
                        for teacher in sorted(section.section_defined_teachers(), key=lambda t: t.number):
                            w.writerow(["add_section_teacher", teacher.number, section.get_teacher_allocation(teacher)])

                w.writerow([])

    # ============================================================================
    # read from CSV file
    # ============================================================================
    @staticmethod
    def read_file(schedule: Schedule, file: str):
        """Parse the details of a schedule from a file, throws an exception if the file cannot be read from"""

        CSVSerializor.last_line_number_read = -1
        CSVSerializor.last_line_read = ""

        with open(file, 'r', newline='') as f:
            reader = csv.reader(f, delimiter=',',
                                quotechar='|', quoting=csv.QUOTE_MINIMAL)
            CSVSerializor._parse_schedule_info(schedule, reader)

        schedule.calculate_conflicts()

    # ============================================================================
    # parse the csv file from an iterable of lists
    # ============================================================================
    @classmethod
    def _parse_schedule_info(cls, schedule: Schedule, reader: Iterable):
        current_obj: Any = None
        course_obj: Optional[Course] = None
        section_obj: Optional[Section] = None
        block_obj: Optional[Block] = None

        for line_number, row in enumerate(reader, start=1):
            CSVSerializor.last_line_number_read = line_number
            CSVSerializor.last_line_read = row

            if not row or row[0] == "":
                continue

            match row[0]:
                # -------------------------------------------------------------
                # collections (teachers, labs, streams
                # -------------------------------------------------------------
                case "lab":
                    (num, descr) = row[1:3]
                    current_obj = schedule.add_update_lab(number=num, description=descr)

                case "unavailable" if current_obj.resource_type == ResourceType.lab:
                    (day, start, duration, movable) = row[1:5]
                    current_obj.add_unavailable_slot(TimeSlot(
                        day=WeekDay.get_from_string(day), start=start, duration=float(duration), movable=bool(int(movable))))

                case 'stream':
                    (number, descr) = row[1:3]
                    current_obj = schedule.add_update_stream(number=number, description=descr)

                case 'teacher':
                    (number, fname, lname, dept, release) = row[1:6]
                    current_obj: Teacher = schedule.add_update_teacher(firstname=fname, lastname=lname, department=dept,
                                                                teacher_id=number)
                    current_obj.release = float(release)

                # -------------------------------------------------------------
                # courses/sections/blocks
                # -------------------------------------------------------------
                case 'course':
                    (number, name, semester, allocation, hours) = row[1:6]
                    course_obj: Course = schedule.add_update_course(number=number,
                                                                    name=name,
                                                                    semester=SemesterType(int(semester)),
                                                                    needs_allocation=bool(int(allocation)),
                                                                    hours = hours)

                case 'section' if course_obj is not None:
                    (number, name, students) = row[2:5]
                    section_obj: Section = course_obj.add_section(number=number, name=name)
                    section_obj.num_students = int(students)

                case "add_stream" if section_obj is not None:
                    s: Optional[Stream] = schedule.get_stream_by_number(row[1])
                    if s is not None:
                        section_obj.add_stream(s)

                case "add_section_teacher" if section_obj is not None:
                    t: Optional[Teacher] = schedule.get_teacher_by_number(row[1])
                    if t is not None:
                        section_obj.add_teacher(t)
                        if len(row) > 2 and row[2] != '':
                            section_obj.set_teacher_allocation(t, float(row[2]))

                case "add_block" if section_obj is not None:
                    (day, start, duration, movable) = row[2:6]
                    x = WeekDay.get_from_string(day)
                    block_obj: Block = section_obj.add_block(
                        day=WeekDay.get_from_string(day), start=float(start), duration=float(duration),
                                 movable=bool(int(movable)))

                case "add_block_teacher" if block_obj is not None:
                    t: Optional[Teacher] = schedule.get_teacher_by_number(row[1])
                    if t is not None:
                        block_obj.add_teacher(t)

                case "add_lab" if block_obj is not None:
                    l: Lab = schedule.get_lab_by_number(row[1])
                    if l is not None:
                        block_obj.add_lab(l)




from __future__ import annotations
import re
import pytest

from src.scheduling_and_allocation.model import Course, Section, WeekDay, ConflictType, Block


@pytest.fixture(scope="module", autouse=True)
def before_and_after_module():
    pass


def init_scenario_and_schedule():
    pass


@pytest.fixture(autouse=True)
def before_and_after():
    pass


class Lab:
    def __lt__(self, other: Lab):
        return True


class Stream:
    def __lt__(self, other: Stream):
        return True


class Teacher:
    def __lt__(self, other: Teacher):
        return True

course = Course()


def test_id():
    """Verifies that the id property works as intended."""
    section = Section(course)
    old_id = section.id
    section = Section(course)
    assert section.id == old_id + 1


def test_id_with_id_given():
    """Verifies that the id property works as intended."""

    existing_id = 912
    section1 = Section(course, section_id=existing_id)
    assert section1.id == existing_id
    section2 = Section(course)
    assert section2.id == existing_id + 1
    section3 = Section(course, section_id=existing_id - 5)
    assert section3.id == existing_id - 5
    section4 = Section(course)
    assert section4.id == section2.id + 1


# region General & Properties
def test_constructor_default_values():
    """Checks that the constructor uses default values"""
    s = Section(course)
    assert isinstance(s.number, str)
    assert isinstance(s.name, str)


def test_section_created_success():
    """Checks that Section is created correctly"""
    number = "3A"

    name = "My Section"
    s = Section(course, number, name)
    assert s.name == name
    assert s.number == number


def test_hours_are_block_hours():
    """Checks if this section has blocks, it returns the number of blocks hours"""
    s = Section(course)
    hours = 22
    s.course.hours = hours
    block1 = s.add_block(WeekDay.Monday, 9.5, 3)
    block2 = s.add_block(WeekDay.Monday, 10.5, 3.5)
    assert s.hours == (block1.duration + block2.duration)


def test_hours_are_course_hours_if_no_blocks():
    """Checks if this section has blocks, if it doesn't, return course hours"""
    s = Section(course)
    hours = 22
    s.course.hours_per_week = hours
    assert s.hours == s.course.hours_per_week


def test_get_title():
    """Checks that title is retrievable and includes name or number"""
    name = "My Section"
    assert name in Section(course, name=name).title

    number = "3A"
    assert number in Section(course, number=number).title


def test_set_num_students():
    """Checks that num_students can be set_default_fonts_and_colours"""
    s = Section(course)
    num = 20
    s.num_students = num
    assert s.num_students == num

def test_is_conflicted_detects_conflicts_correctly():
    """Checks that the is_conflicted method correctly picks up conflicted_number blocks"""
    s = Section(course)
    block1 = s.add_block(WeekDay.Monday, 10.5, 3)
    block1.add_conflict(ConflictType.LUNCH)
    assert s.is_conflicted()


def test_is_conflicted_detects_ok_correctly():
    """Checks that the is_conflicted method doesn't return false positive"""
    s = Section(course)
    s.add_block(WeekDay.Monday, 13, 2)
    s.add_block(WeekDay.Monday, 13, 2)
    s.add_block(WeekDay.Monday, 13, 2)
    assert not s.is_conflicted()


def test_string_representation_with_valid_name():
    n = "10"
    s = Section(course, number=n, name="For students taking calculus")
    assert re.search(f"{n}" + r".*?For students taking calculus", str(s))


def test_string_representation_with_invalid_name():
    n = "10"
    s = Section(course, number=n, name=f"Section {n}")
    assert str(s) == f"Section {10}"


def test_string_representation_with_no_name():
    n = "10"
    s = Section(course, number=n)
    assert str(s) == f"Section {10}"


# endregion

# region Block
def test_add_block_valid():
    """Checks that a valid blocks can be added"""
    s = Section(course)
    b = s.add_block(WeekDay.Monday, 13, 2)
    assert b in s.blocks()


def test_get_block_by_id_valid():
    """Checks that blocks can be retrieved by id"""
    s = Section(course)
    b1 = s.add_block(WeekDay.Monday, 13, 2)
    b1_id = b1.id
    s.add_block(WeekDay.Tuesday, 12, 5)
    assert s.get_block_by_id(b1_id) == b1


def test_get_block_by_id_invalid():
    """Checks that None is returned when get_block_by_id is passed a non-blocks id"""
    s = Section(course)
    assert s.get_block_by_id(-1) is None


def test_remove_block_valid():
    """Checks that when passed a valid blocks, it will be removed"""
    s = Section(course)
    b1 = s.add_block(WeekDay.Monday, 13, 2)
    b2 = s.add_block(WeekDay.Monday, 13, 2)
    s.remove_block(b1)
    assert b1 not in s.blocks()
    assert b2 in s.blocks()


# endregion

# region Lab

def test_assign_lab_valid():
    """Checks that a valid lab can be added, if there is a block"""
    s = Section(course)
    s.add_block(WeekDay.Monday, 13, 2)
    lab = Lab()
    s.add_lab(lab)
    assert lab in s.labs()


def test_remove_lab_valid():
    """Checks that when passed a valid lab, it will be removed & deleted"""
    s = Section(course)
    s.add_block(WeekDay.Monday, 13, 2)
    lab = Lab()
    s.add_lab(lab)
    s.remove_lab(lab)
    assert lab not in s.labs()


def test_remove_lab_not_there():
    """Checks that when passed not-included lab, it will still be 'removed' """
    s = Section(course)
    s.add_block(WeekDay.Monday, 13, 2)
    lab = Lab()
    s.remove_lab(lab)
    assert lab not in s.labs()


# endregion

# region Teacher

def test_assign_teacher_valid():
    """Checks that a valid teacher can be added"""
    s = Section(course)
    s.add_block()
    t = Teacher()
    s.add_teacher(t)
    assert t in s.teachers()


def test_get_teacher_allocation_valid():
    """Checks that get_teacher_allocation returns the correct allocation hours"""
    s = Section(course)
    t = Teacher()
    hours = 12
    s.add_teacher(t)
    s.set_teacher_allocation(t, hours)
    assert s.get_teacher_allocation(t) == hours


def test_assign_teacher_valid_allocation_hours_default_to_section_hours():
    """Checks that a valid teacher can be added, and that the teacher allocation hours is set_default_fonts_and_colours to section hours"""
    s = Section(course)
    s.add_block()
    course.hours = 15
    t = Teacher()
    s.add_teacher(t)
    assert s.get_teacher_allocation(t) == s.hours
    t2 = Teacher()
    s.add_teacher(t2)
    assert s.get_teacher_allocation(t2) == s.hours

def test_has_teacher_valid():
    """Checks has_teacher returns True if teacher is included"""
    s = Section(course)
    s.add_block()
    t = Teacher()
    s.add_teacher(t)
    assert s.has_teacher(t)


def test_has_teacher_not_found():
    """Checks has_teacher returns False if teacher is not included"""
    s = Section(course)
    s.add_block()
    t = Teacher()
    assert not s.has_teacher(t)

def test_remove_all_deletes_all_teachers():
    """Checks that remove_all_teachers will correctly delete them all"""
    s = Section(course)
    s.add_teacher(Teacher())
    s.add_teacher(Teacher())
    s.add_teacher(Teacher())
    s.remove_all_teachers()
    assert len( s.teachers()) == 0


def test_teachers_in_blocks_in_teachers_property():
    s = Section(course)
    b1: Block = s.add_block(WeekDay.Monday, 13, 1.5)
    b2: Block = s.add_block(WeekDay.Monday, 13, 1.5)
    t1 = Teacher()
    t2 = Teacher()
    t3 = Teacher()
    s.add_teacher(t3)
    s.add_teacher(t1)
    s.add_teacher(t2)

    assert len(s.teachers()) == 3
    assert t1 in s.teachers()
    assert t2 in s.teachers()
    assert t3 in s.teachers()
    assert t1 in b1.teachers()
    assert t3 in b1.teachers()
    assert t2 in b2.teachers()
    assert t3 in b2.teachers()
    assert len(b1.teachers()) == 3
    assert len(b2.teachers()) == 3

def test_remove_teacher_valid():
    """Checks that when passed a valid teacher, it will be removed"""
    s = Section(course)
    s.add_block()
    t = Teacher()
    t2 = Teacher()
    s.add_teacher(t)
    s.add_teacher(t2)
    s.remove_teacher(t)
    assert not s.has_teacher(t)
    assert s.has_teacher(t2)


def test_remove_all_deletes_all_teachers():
    """Checks that remove_all_teachers will correctly delete them all"""
    s = Section(course)
    s.add_block()
    s.add_teacher(Teacher())
    s.add_teacher(Teacher())
    s.add_teacher(Teacher())
    s.remove_all_teachers()
    assert len( s.teachers()) == 0



# -------------------------------------------------------------------------------------------------
# teacher allocation
# -------------------------------------------------------------------------------------------------
def test_section_allocation_no_blocks():
    """Checks that a valid teacher can be added, and that the teacher allocation hours is saved as set"""
    s = Section(course)
    t = Teacher()
    s.set_teacher_allocation(t, 3)
    assert s.get_teacher_allocation(t) == 3

def test_section_allocation_hrs_equals_block_time():
    """If allocation is equal to total block time, then assign teacher to all blocks"""
    s = Section(course)
    b1 = s.add_block(duration=5)
    b2 = s.add_block(duration=2)
    t = Teacher()
    s.set_teacher_allocation(t, 7)
    assert s.get_teacher_allocation(t) == 7
    assert b1.has_teacher(t)
    assert b2.has_teacher(t)

def test_section_allocation_hrs_fits_one_block_time():
    """If allocation is equal to total block time, then assign teacher to all blocks"""
    s = Section(course)
    b1 = s.add_block(duration=5)
    b2 = s.add_block(duration=2)
    t = Teacher()
    s.set_teacher_allocation(t, 5)
    assert s.get_teacher_allocation(t) == 5
    assert b1.has_teacher(t)
    assert not b2.has_teacher(t)

def test_section_allocation_hrs_does_not_any_block_time():
    """If allocation cannot fit properly into combination of blocks, no blocks are set"""
    s = Section(course)
    b1 = s.add_block(duration=5)
    b2 = s.add_block(duration=2)
    t = Teacher()
    s.set_teacher_allocation(t, 1.5)
    assert s.get_teacher_allocation(t) == 1.5
    assert not b1.has_teacher(t)
    assert not b2.has_teacher(t)

def test_section_allocation_hrs_fits_one_block_time_with_leftover():
    """If allocation cannot fit properly into combination of blocks, no blocks are set"""
    s = Section(course)
    b1 = s.add_block(duration=5)
    b2 = s.add_block(duration=2)
    t = Teacher()
    s.set_teacher_allocation(t, 3)
    assert s.get_teacher_allocation(t) == 3
    assert not b1.has_teacher(t)
    assert not b2.has_teacher(t)

def test_section_allocation_hrs_larger_than_sum_block_times():
    """If allocation cannot fit properly into combination of blocks, no blocks are set"""
    s = Section(course)
    b1 = s.add_block(duration=5)
    b2 = s.add_block(duration=2)
    t = Teacher()
    s.set_teacher_allocation(t, 8)
    assert s.get_teacher_allocation(t) == 8
    assert not b1.has_teacher(t)
    assert not b2.has_teacher(t)

def test_section_allocation_preference_given_to_block_with_no_teacher():
    """When assigning blocks, first assign to blocks that have no teacher"""
    s = Section(course)
    b1 = s.add_block(duration=2)
    b2 = s.add_block(duration=2)
    t1 = Teacher()
    t2 = Teacher()
    b1.add_teacher(t2)
    s.set_teacher_allocation(t1, 2)
    assert s.get_teacher_allocation(t1) == 2
    assert not b1.has_teacher(t1)
    assert b2.has_teacher(t1)

def test_section_allocation_preference_given_to_block_with_no_teacher2():
    """When assigning blocks, first assign to blocks that have no teacher"""
    s = Section(course)
    b1 = s.add_block(duration=2)
    b2 = s.add_block(duration=2)
    t1 = Teacher()
    t2 = Teacher()
    b2.add_teacher(t2)
    s.set_teacher_allocation(t1, 2)
    assert s.get_teacher_allocation(t1) == 2
    assert not b2.has_teacher(t1)
    assert b1.has_teacher(t1)

def test_section_allocation_assign_to_block_with_teacher():
    """When assigning blocks, if not possible otherwise, assign to block with teacher"""
    s = Section(course)
    b1 = s.add_block(duration=2)
    b2 = s.add_block(duration=2)
    t1 = Teacher()
    t2 = Teacher()
    s.add_teacher(t2)
    s.set_teacher_allocation(t1, 2)
    assert s.get_teacher_allocation(t1) == 2
    assert ((not b2.has_teacher(t1) and b1.has_teacher(t1)) or
            (b2.has_teacher(t1) and not b1.has_teacher(t1)))

def test_section_allocation_hours_returned_even_if_teacher_assigned_to_section_without_allocation():
    """When assigning blocks, first assign to blocks that have no teacher"""
    s = Section(course)
    b1 = s.add_block(duration=2)
    b2 = s.add_block(duration=2)
    t2 = Teacher()
    s.add_teacher(t2)
    assert s.get_teacher_allocation(t2) == 4

def test_has_allocated_teacher_valid_false_unless_specifically_allocated():
    s = Section(course)
    s.add_block()
    t = Teacher()
    s.add_teacher(t)
    assert s.has_teacher(t)
    assert t not in s.section_defined_teachers()

def test_has_allocated_teacher_valid_true_if_specifically_allocated():
    s = Section(course)
    s.add_block()
    t = Teacher()
    s.set_teacher_allocation(t, 5)
    assert not s.has_teacher(t)
    assert t in s.section_defined_teachers()
    assert s.has_allocated_teacher(t)


def test_remove_allocation_teacher_valid1():
    """Remove teacher removes allocated teacher (whose hrs fit into blocks properly)"""
    s = Section(course)
    hours = 5
    s.add_block(3)
    s.add_block(2)
    t_added_teacher = Teacher()
    t_allocated_teacher = Teacher()
    s.add_teacher(t_added_teacher)
    s.set_teacher_allocation(t_allocated_teacher, hours)
    s.remove_allocation(t_allocated_teacher)
    assert s.get_teacher_allocation(t_allocated_teacher) == 0
    assert not s.has_teacher(t_allocated_teacher)
    assert not s.has_allocated_teacher(t_allocated_teacher)
    assert s.has_teacher(t_added_teacher)

def test_remove_teacher_valid2():
    """Remove teacher removes allocated teacher (whose hrs don't fit into blocks properly)"""
    s = Section(course)
    hours = 7
    s.add_block(3)
    s.add_block(2)
    t_added_teacher = Teacher()
    t_allocated_teacher = Teacher()
    s.add_teacher(t_added_teacher)
    s.set_teacher_allocation(t_allocated_teacher, hours)
    s.remove_allocation(t_allocated_teacher)
    assert s.get_teacher_allocation(t_allocated_teacher) == 0
    assert not s.has_teacher(t_allocated_teacher)
    assert not s.has_allocated_teacher(t_allocated_teacher)
    assert s.has_teacher(t_added_teacher)


def test_remove_teacher_not_remove_allocated_teacher():
    """Checks that when passed a valid allocated teacher, is not removed by simple remove """
    """Remove teacher removes allocated teacher (whose hrs don't fit into blocks properly)"""
    s = Section(course)
    hours = 7
    s.add_block(3)
    s.add_block(2)
    t_added_teacher = Teacher()
    t_allocated_teacher = Teacher()
    s.add_teacher(t_added_teacher)
    s.set_teacher_allocation(t_allocated_teacher, hours)
    s.remove_teacher(t_allocated_teacher)
    assert s.get_teacher_allocation(t_allocated_teacher) == 7
    assert not s.has_teacher(t_allocated_teacher)
    assert s.has_allocated_teacher(t_allocated_teacher)
    assert s.has_teacher(t_added_teacher)



# -------------------------------------------------------------------------------------------------
# stream
# -------------------------------------------------------------------------------------------------

def test_assign_stream_valid():
    """Checks that a valid stream can be added"""
    s = Section(course)
    st = Stream()
    s.add_stream(st)
    assert st in s.streams()


def test_has_stream_valid():
    """Checks has_stream_with_id returns True if stream is included"""
    s = Section(course)
    st = Stream()
    s.add_stream(st)
    assert s.has_stream(st)


def test_has_stream_not_found():
    """Checks has_stream_with_id returns False if stream is not included"""
    s = Section(course)
    st = Stream()
    assert not s.has_stream(st)


def test_remove_stream_valid():
    """Checks that when passed a valid stream"""
    s = Section(course)
    st = Stream()
    s.add_stream(st)
    s.remove_stream(st)
    assert st not in s.streams()


def test_remove_stream_not_there():
    """Checks that when passed not-included stream, it will still be 'removed' """
    s = Section(course)
    st = Stream()
    s.remove_stream(st)
    assert s not in s.streams()


def test_remove_all_deletes_all_streams():
    """Checks that remove_all_streams will correctly delete them all"""
    s = Section(course)
    s.add_stream(Stream())
    s.add_stream(Stream())
    s.add_stream(Stream())
    s.remove_all_streams()
    assert not s.streams()

def test_remove_teacher_from_all_blocks_removes_from_section():
    s=Section(course)
    teacher=Teacher()
    b=s.add_block(1,1,1)
    b2=s.add_block(1,1,1)
    s.add_teacher(teacher)
    assert teacher in b2.teachers()
    assert teacher in b.teachers()

    b.remove_teacher(teacher)
    b2.remove_teacher(teacher)
    assert len(s.teachers()) == 0



# endregion



from os import path
from typing import Callable, Any, Optional

import pytest

from src.scheduling_and_allocation.gui_pages import EditResourcesTk, DEColumnDescription
from src.scheduling_and_allocation.model import Schedule, ResourceType
from src.scheduling_and_allocation.presenter.edit_resources import EditResources

SCHEDULE_FILE = path.dirname(__file__) + "/test.csv"
BIN_DIR = path.dirname(__file__) + "/../../schedule"

# ============================================================================
# overload the gui
# ============================================================================
CURRENT_TEST_FILENAME = SCHEDULE_FILE


class EditResourcesTkTest(EditResourcesTk):
    def __init__(self,
                 parent,
                 event_delete_handler: Callable[[list[str], ...], None] = lambda x, *_: None,
                 event_save_handler: Callable[[...], None] = lambda *_: None,
                 colours=None,
                 ):
        self.frame = parent
        self.data_entry: list = list()
        self.delete_handler = event_delete_handler
        self.save_handler = event_save_handler

    def initialize_columns(self, column_descriptions: list[DEColumnDescription], disabled=[]):
        self.called_initialize_columns = True
        self.column_descriptions = column_descriptions

    def refresh(self, data: list[list[Any]]):
        self.called_refresh = True
        self.data_entry = data

    def get_all_data(self) -> list[list[str]]:
        self.called_get_all_data = True
        return self.data_entry


# ============================================================================
# tests
# ============================================================================
@pytest.fixture()
def gui():
    return EditResourcesTkTest(None, None, None)


dirty_flag = False


@pytest.fixture()
def dirty():
    global dirty_flag
    dirty_flag = False

    def set_dirty(value: Optional[bool] = None) -> bool:
        global dirty_flag
        if value is not None:
            dirty_flag = value
        return dirty_flag

    return set_dirty


@pytest.fixture()
def schedule_obj():
    schedule = Schedule()

    # teachers
    schedule.add_update_teacher("Jane", "Doe", "0.25")
    schedule.add_update_teacher("John", "Doe")
    schedule.add_update_teacher("Babe", "Ruth")

    # labs
    schedule.add_update_lab("P107", "C-Lab")
    schedule.add_update_lab("P322", "Mac Lab")
    schedule.add_update_lab("P325")

    # streams
    schedule.add_update_stream("1A", "calculus available")
    schedule.add_update_stream("1B", "no calculus")
    schedule.add_update_stream("2A")
    schedule.add_update_stream("2B")

    return schedule


def test_init_teacher(gui, dirty, schedule_obj):
    """init
    - initializes columns
    """
    # execute
    EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # verify
    assert gui.called_initialize_columns
    columns = gui.column_descriptions
    assert columns[1].property == 'firstname'
    assert columns[2].property == 'lastname'
    assert columns[3].property == 'release'


def test_init_stream(gui, dirty, schedule_obj):
    """init
    - initializes columns
    """
    # execute
    EditResources(dirty, None, ResourceType.stream, schedule_obj, gui)

    # verify
    assert gui.called_initialize_columns
    columns = gui.column_descriptions
    assert columns[0].property == 'number'
    assert columns[1].property == 'description'


def test_init_lab(gui, dirty, schedule_obj):
    """init
    - initializes columns
    """
    # execute
    EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)

    # verify
    assert gui.called_initialize_columns
    columns = gui.column_descriptions
    assert columns[0].property == 'number'
    assert columns[1].property == 'description'


def test_init_no_change_dirty1(gui, dirty, schedule_obj):
    """init
    - no change in dirty flag
    """

    # prepare
    dirty_set_to = True
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to

    # execute
    EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)

    # verify
    assert dirty_flag == dirty_set_to


def test_init_no_change_dirty2(gui, dirty, schedule_obj):
    """init
    - no change in dirty flag
    """

    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to

    # execute
    EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)

    # verify
    assert dirty_flag == dirty_set_to


def test_refresh_no_data(gui, dirty, schedule_obj):
    """refresh data
    - calls gui refresh
    - purges delete queue
    """
    # prepare
    schedule_obj = Schedule()
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.refresh()

    # verify
    assert gui.called_refresh
    assert len(gui.data_entry) == 0
    assert len(obj.delete_queue) == 0


def test_refresh_with_data(gui, dirty, schedule_obj):
    """refresh data
    - calls gui refresh
    - purges delete queue
    """
    # prepare
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.refresh()

    # verify
    assert gui.called_refresh
    assert len(gui.data_entry) == 3
    assert len(obj.delete_queue) == 0


def test_refresh_no_change_dirty1(gui, dirty, schedule_obj):
    """refresh data
    - no change dirty flag
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.refresh()

    # verify
    assert dirty_flag == dirty_set_to


def test_refresh_no_change_dirty2(gui, dirty, schedule_obj):
    """refresh data
    - no change dirty flag
    """
    # prepare
    dirty_set_to = True
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.refresh()

    # verify
    assert dirty_flag == dirty_set_to


def test_adding_new_teacher(gui, dirty, schedule_obj):
    """save after teacher added
    - teacher is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append(["", "New", "Teacher", ".21"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.teachers()) == 4
    new_teacher = schedule_obj.get_teacher_by_name("New", "Teacher")
    assert new_teacher is not None
    assert new_teacher.release == 0.21
    assert dirty_flag


def test_adding_new_teacher_no_release_specified(gui, dirty, schedule_obj):
    """save after teacher added
    - teacher is added to schedule, but use default release
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append([ "", "New", "Teacher"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.teachers()) == 4
    new_teacher = schedule_obj.get_teacher_by_name("New", "Teacher")
    assert new_teacher is not None
    assert dirty_flag

def test_modifying_teacher_not_new_obj(gui, dirty, schedule_obj):
    """save after teacher release modified
    - teacher modification is added to schedule, without changing the actual object
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    teacher_obj_id = id(schedule_obj.get_teacher_by_number("Doe_Jane"))
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    obj.refresh()
    gui.data_entry[1][3] = "3"

    # execute
    obj.save(gui.data_entry)

    # validate
    teacher = schedule_obj.get_teacher_by_number("Doe_Jane")
    assert teacher is not None
    assert teacher.firstname == "Jane"
    assert teacher.release == 3.0
    assert id(teacher) == teacher_obj_id
    assert dirty_flag



def test_modifying_teacher(gui, dirty, schedule_obj):
    """save after teacher modified
    - teacher modification is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    obj.refresh()
    gui.data_entry[1][2] = "new name"
    teacher_id = gui.data_entry[1][0]

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.teachers()) == 3
    teacher = schedule_obj.get_teacher_by_number(teacher_id)
    assert teacher is not None
    assert teacher.firstname == "Jane"
    assert teacher.lastname == "new name"
    assert dirty_flag


def test_adding_new_lab(gui, dirty, schedule_obj):
    """save after lab added
    - lab is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append(["P000", "New"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.labs()) == 4
    new_lab = schedule_obj.get_lab_by_number("P000")
    assert new_lab is not None
    assert dirty_flag


def test_adding_new_lab2(gui, dirty, schedule_obj):
    """save after lab added, not specifying description
    - lab is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append(["P000"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.labs()) == 4
    new_lab = schedule_obj.get_lab_by_number("P000")
    assert new_lab is not None
    assert dirty_flag


def test_modifying_lab(gui, dirty, schedule_obj):
    """save after lab modified
    - lab modification is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    gui.data_entry[1][1] = "new name"
    lab_id = gui.data_entry[1][0]

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.labs()) == 3
    lab = schedule_obj.get_lab_by_number(lab_id)
    assert lab is not None
    assert lab.description == "new name"
    assert dirty_flag

def test_modifying_lab_not_new_obj(gui, dirty, schedule_obj):
    """save after lab description modified
    - lab modification is added to schedule, without changing the actual object
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    lab_obj_id = id(schedule_obj.get_lab_by_number("P107"))
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    gui.data_entry[0][1] = "something else"

    # execute
    obj.save(gui.data_entry)

    # validate
    lab = schedule_obj.get_lab_by_number("P107")
    assert lab is not None
    assert lab.description == "something else"
    assert id(lab) == lab_obj_id
    assert dirty_flag


def test_adding_new_stream(gui, dirty, schedule_obj):
    """save after stream added
    - lab is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.stream, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append(["1C", "New"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.streams()) == 5
    new_stream = schedule_obj.get_stream_by_number("1C")
    assert new_stream is not None
    assert dirty_flag


def test_adding_new_stream2(gui, dirty, schedule_obj):
    """save after stream added, not specifying description
    - lab is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.stream, schedule_obj, gui)
    obj.refresh()
    gui.data_entry.append(["1C"])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.streams()) == 5
    new_stream = schedule_obj.get_stream_by_number("1C")
    assert new_stream is not None
    assert dirty_flag


def test_modifying_stream(gui, dirty, schedule_obj):
    """save after stream modified
    - lab modification is added to schedule
    - dirty flag is set
    """
    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    obj = EditResources(dirty, None, ResourceType.stream, schedule_obj, gui)
    obj.refresh()
    gui.data_entry[1][1] = "new name"
    stream_id = gui.data_entry[1][0]

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.streams()) == 4
    stream = schedule_obj.get_stream_by_number(stream_id)
    assert stream is not None
    assert stream.description == "new name"
    assert dirty_flag


def test_save_with_no_changes1(gui, dirty, schedule_obj):
    """no changes, then save
    - dirty flag unchanged
    """
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.save(gui.data_entry)

    # validate
    assert dirty_flag == dirty_set_to


def test_save_with_no_changes2(gui, dirty, schedule_obj):
    """no changes, then save
    - dirty flag unchanged
    """
    dirty_set_to = True
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)

    # execute
    obj.save(gui.data_entry)

    # validate
    assert dirty_flag == dirty_set_to


def test_save_with_handle_empty_rows1(gui, dirty, schedule_obj):
    """add empty rows
    - no changes
    - dirty flag unchanged
    """

    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    gui.data_entry.append(["", "", "", ""])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.streams()) == 4
    assert dirty_flag == dirty_set_to


def test_save_with_handle_empty_rows2(gui, dirty, schedule_obj):
    """add empty rows
    - no changes
    - dirty flag unchanged
    """

    # prepare
    dirty_set_to = True
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    gui.data_entry.append(["", "", "", ""])

    # execute
    obj.save(gui.data_entry)

    # validate
    assert len(schedule_obj.streams()) == 4
    assert dirty_flag == dirty_set_to


def test_delete_teacher(gui, dirty, schedule_obj):
    """delete teacher
    - teacher is removed
    - dirty flag is set
    """

    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.teacher, schedule_obj, gui)
    obj.refresh()
    teacher_id = gui.data_entry[1][0]

    # execute
    obj.delete_obj(gui.data_entry[1])
    obj.save(gui.data_entry)

    # validate
    assert obj.schedule.get_teacher_by_number(teacher_id) is None
    assert dirty_flag


def test_delete_lab(gui, dirty, schedule_obj):
    """delete lab
    - lab is removed
    - dirty flag is set
    """

    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    lab_id = gui.data_entry[1][0]

    # execute
    obj.delete_obj(gui.data_entry[1])
    obj.save(gui.data_entry)

    # validate
    assert obj.schedule.get_lab_by_number(lab_id) is None
    assert dirty_flag


def test_delete_stream(gui, dirty, schedule_obj):
    """delete stream
    - stream is removed
    - dirty flag is set
    """

    # prepare
    dirty_set_to = False
    dirty(dirty_set_to)
    assert dirty_flag == dirty_set_to
    obj = EditResources(dirty, None, ResourceType.stream, schedule_obj, gui)
    obj.refresh()
    stream_id = gui.data_entry[1][0]

    # execute
    obj.delete_obj(gui.data_entry[1])
    obj.save(gui.data_entry)

    # validate
    assert obj.schedule.get_stream_by_number(stream_id) is None
    assert dirty_flag


def test_delete_happens_before_resource_update(gui, dirty, schedule_obj):
    """delete obj with id 'a', add new obj with id 'a'
    - an object with id 'a' exists, and has new properties
    Reasoning: if resource updates first, then deletes, the object will be deleted and no longer exist,
               if resource deletes first, then updates, new object will be created
    """

    # prepare
    obj = EditResources(dirty, None, ResourceType.lab, schedule_obj, gui)
    obj.refresh()
    lab_id = gui.data_entry[1][0]

    # execute
    obj.delete_obj(gui.data_entry[1])
    gui.data_entry.append([lab_id, "new object"])
    obj.save(gui.data_entry)

    # validate
    assert obj.schedule.get_lab_by_number(lab_id) is not None
    lab = obj.schedule.get_lab_by_number(lab_id)
    assert lab.description == "new object"



from dataclasses import dataclass
from tkinter import *

from src.scheduling_and_allocation.gui_pages.view_canvas_tk import ViewCanvasTk
from src.scheduling_and_allocation.model import ResourceType

mw = Tk()
cn = Canvas(mw, height=700, width=700, background="white")
cn.pack()

vc = ViewCanvasTk(cn)

@dataclass
class ExampleBlock:
    start_time: float = 8.0
    day: int = 1
    duration: float = 1.5
    resource_type: ResourceType = ResourceType.teacher
bl1 = ExampleBlock()
bl2 = ExampleBlock(start_time=9.5, day=1)
vc.draw_block(ResourceType.teacher, bl1.day, bl1.start_time, bl1.duration,"Block1","g01",movable=True)
vc.draw_block(ResourceType.teacher, bl2.day, bl2.start_time, bl2.duration,"Block2","g02", movable=True)


mw.mainloop()




from .InitGuiFontsAndColours import set_default_fonts_and_colours, TkColours, TkFonts, get_fonts_and_colours
from .InitGuiFontsAndColours import colours as tk_colours
from .InitGuiFontsAndColours import fonts as tk_fonts

from .AdvancedTreeview import AdvancedTreeview
from .DragNDrop import DragNDropManager
from .Scrolled import Scrolled
from .idlelib_tooltip import Hovertip
from .Pane import Pane
from .ToolBar import ToolBar

from . import InitGuiFontsAndColours





from tkinter import *
import os
import sys

from src.scheduling_and_allocation.gui_dialogs.add_edit_block_dialog_tk import AddEditBlockDialogTk
from src.scheduling_and_allocation.model import Schedule

mw=Tk()
frame = Frame(mw)
frame.pack()
mw.geometry("400x400")

schedule = Schedule()

# teachers
t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id="4")

# labs
l1 = schedule.add_update_lab("P107", "C-Lab")
l2 = schedule.add_update_lab("P322", "Mac Lab")
l3 = schedule.add_update_lab("P325")
l4 = schedule.add_update_lab("BH311", "Britain Hall")
def apply_changes(number, hour, teachers, labs):
    print(number, hour, teachers, labs)

def go_edit():
    db = AddEditBlockDialogTk(frame, "edit", 1.5, [t1, t2, t3], [t4], [l2], [l1, l3, l4], apply_changes)
def go_add():
    db = AddEditBlockDialogTk(frame, "add", 1.5, [t1, t2, t3], [t4], [l2], [l1, l3, l4], apply_changes)

Button(frame, text="Edit Class Time", command=go_edit).pack()
Button(frame, text="Add Class Times", command=go_add).pack()

mw.mainloop()


from typing import Literal
import tkinter as tk
from ..modified_tk import Scrolled

SCROLLBAR_DIR = Literal["s", "w", "e", "n", "se", "sw", "ne", "nw"]

# =====================================================================================================================
# create some text info that is read only
# =====================================================================================================================
class ReadOnlyTextTk:
    def __init__(self, frame, text: list[str] = None,
                 height: int = 20, width: int = 50,
                 scrollbars: SCROLLBAR_DIR = 'se',
                 wrap="none"):
        """
        :param frame: the container where the text widget will be placed
        :param text: the text to display (list)
        :param height: height of text box
        :param width: width of text box
        :param scrollbars: scrollbars ?
        :param wrap: do you want the text to wrap-around?
        """

        # remove any pre-existing widgets in this frame
        for w in frame.winfo_children():
            w.destroy()

        s: Scrolled = Scrolled(
            frame,
            'Text',
            height=height,
            width=width,
            scrollbars=scrollbars,
            wrap=wrap
        )
        s.pack(expand=True, fill="both")
        self.textbox: tk.Text = s.widget
        if text is not None:
            self.write(text)
        self.textbox.config(state="disabled")

    def write(self, text):
        self.textbox.config(state="normal")
        self.textbox.delete("1.0", "end")
        for txt in text:
            self.textbox.insert('end', txt + "\n")
        self.textbox.config(state="disabled")




"""Entry point for the GUI_Pages Allocation Management Tool"""
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Optional

from .allocation_editor import AllocationEditor
from .edit_courses import EditCourses
from .edit_resources import EditResources
from .menus_main_menu_allocation import set_menu_event_handler_allocation, main_menu_allocation
from .student_numbers import StudentNumbers

from ..Utilities import Preferences
from ..gui_pages.allocation_manager_tk import AllocationManagerTk, set_main_page_event_handler
from ..model import SemesterType, Schedule, ResourceType, CouldNotReadFileError

# =====================================================================================
# Notebook book-keeping
# =====================================================================================
@dataclass
class NBTabInfo:
        name: str
        label: str
        subpages: list = field(default_factory=list)
        frame_args: dict[str, Any] =  field(default_factory=dict)
        creation_handler: Callable = lambda *_: None
        is_default_page: bool = False

DIRECTORY = str
PATH = str
VALID_SEMESTERS = [SemesterType.fall, SemesterType.winter]


# =====================================================================================
# Allocation Manager
# =====================================================================================
class AllocationManager:
    NB_course = "Courses"
    NB_teacher = "Teachers"
    NB_students = "Students"
    NB_allocation = "Allocation"


    def __init__(self, bin_dir: DIRECTORY, gui: Optional[AllocationManagerTk] = None):

        self.preferences: Preferences = Preferences()
        self.schedules: dict[SemesterType, Optional[Schedule]] = {s:None for s in VALID_SEMESTERS}
        self._previous_filenames: dict[SemesterType, str] = {s:"" for s in VALID_SEMESTERS}
        self._dirty_flag = False
        self.current_tab: Optional[str] = None
        self.standard_page = None

        # gui is optional so that we can test the presenter more readily
        if gui:
            self.gui = gui
        else:
            self.gui: AllocationManager = AllocationManagerTk('Allocation', self.preferences, bin_dir)

        self._schedule_filenames: dict[SemesterType, PATH] = {s:"" for s in VALID_SEMESTERS}

        self._allocation_manager_already_open = False

        # --------------------------------------------------------------------
        # required notebook pages
        # --------------------------------------------------------------------
        course_tab = {}
        teacher_tab = {}
        student_tab = {}
        allocation_tab = {}
        for semester in VALID_SEMESTERS:
            allocation_tab[semester] = NBTabInfo(label=f"{semester.name} {self.NB_allocation}",
                                             name=f"{semester.name} {self.NB_allocation}")
            course_tab[semester] = NBTabInfo(label=f"{semester.name} {self.NB_course}",
                                             name=f"{semester.name} {self.NB_course}")
            teacher_tab[semester] = NBTabInfo(label=f"{semester.name} {self.NB_teacher}",
                                             name=f"{semester.name} {self.NB_teacher}")
            student_tab[semester] = NBTabInfo(label=f"{semester.name} {self.NB_students}",
                                             name=f"{semester.name} {self.NB_students}")

        self._notebook_tabs = []
        for semester in VALID_SEMESTERS:
            self._notebook_tabs.append(
                NBTabInfo(label=semester.name, name=semester.name,
                      subpages=[allocation_tab[semester], course_tab[semester],
                                teacher_tab[semester], student_tab[semester]]
                          )
            )

        # --------------------------------------------------------------------
        # create the Menu and Toolbars
        # --------------------------------------------------------------------
        set_menu_event_handler_allocation("file_new", self.new_menu_event)
        set_menu_event_handler_allocation("file_open", self.open_menu_event)
        set_menu_event_handler_allocation("file_save", self.save_schedule)
        set_menu_event_handler_allocation("file_exit", self.menu_exit_event)

        self.gui.toggle_auto_save = self.auto_save_set

        set_main_page_event_handler("go", self.go)
        set_main_page_event_handler("exit", self.exit_event)

        (self._toolbar_buttons, self._button_properties, self._menu) = main_menu_allocation(VALID_SEMESTERS)

        # --------------------------------------------------------------------
        # create the Gui Main Window
        # --------------------------------------------------------------------
        self.gui.create_menu_and_toolbars(self._toolbar_buttons, self._button_properties, self._menu)
        self.gui.create_welcome_page(VALID_SEMESTERS)
        self.gui.create_status_bar()
        self.gui.notebook_tab_changed_handler = self.notebook_tab_has_changed

        self._previous_filenames: dict[SemesterType, PATH] = {}
        for semester in VALID_SEMESTERS:
            self.preferences.semester(semester.name)
            self.previous_filename(semester, self.preferences.previous_file())

        self.set_dirty_indicator()
        self.gui.start_event_loop()


    # ============================================================================================
    #  getters/setters - filenames
    # ============================================================================================
    def schedule_filename(self, semester: SemesterType=SemesterType.any, value=None) -> str:
        if value is not None:
            self._schedule_filenames[semester] = value
            self.gui.schedule_filename(semester,value)
            if value != "":
                self.previous_filename(semester,value)
        return self._schedule_filenames[semester]


    def previous_filename(self, semester:SemesterType=SemesterType.any, value=None)->str:
        if value is not None:
            self.preferences.semester(semester.name)
            self.preferences.previous_file(value)
            self._previous_filenames[semester] = value
            self.preferences.save()

        return self._previous_filenames.get(semester,"")

    # ============================================================================================
    # Properties - is data changed (dirty)
    # ============================================================================================
    @property
    def dirty_flag(self) -> bool:
        """is the data different from what was saved on disk?"""
        return self._dirty_flag

    @dirty_flag.setter
    def dirty_flag(self, value):
        self._dirty_flag = value
        self.set_dirty_indicator()

    # ============================================================================================
    # Events ... open
    # ============================================================================================
    def open_menu_event(self, semester):
        """open a file"""
        self.open_menu_event_from_main_page(semester)
        self.gui.create_standard_page(self._notebook_tabs, reset=True)

    def open_menu_event_from_main_page(self, semester):
        """open a file"""
        self.preferences.semester(semester.name)
        filename = self.gui.select_file_to_open(f"Open Schedule ({semester.name.upper()})")
        self._open_file(filename, semester)

    def open_previous_file_event(self, semester):
        """open previously opened file"""
        self.open_previous_file_event_from_main_page(semester)
        self.gui.create_standard_page(self._notebook_tabs, reset=True)

    def open_previous_file_event_from_main_page(self, semester):
        """open previously opened file"""
        self.preferences.semester(semester.name)
        filename = self.preferences.previous_file()
        self._open_file(filename, semester)

    def _open_file(self, filename: str, semester):
        """generic open file method"""
        if filename:
            try:
                schedule = Schedule(filename)
                self._allocation_manager_already_open = False
                self.schedules[semester] = schedule
                self.schedule_filename(semester, filename)
                self.dirty_flag = False
                if self.standard_page is not None:
                    self.gui.create_standard_page(self._notebook_tabs)


            except CouldNotReadFileError as e:
                self.gui.show_error("Read File", str(e))

    # ============================================================================================
    # Event handlers - new
    # ============================================================================================
    def new_menu_event(self, semester: SemesterType):
        """create a new file"""
        schedule = Schedule()
        self.schedules[semester] = schedule
        self.schedule_filename(semester, "")
        self.dirty_flag = True
        self.gui.create_standard_page(self._notebook_tabs, reset=True)

    # ============================================================================================
    # Event handlers - save
    # ============================================================================================
    def save_schedule(self, *_):
        """generic save file method"""

        for semester in self.schedules.keys():
            if self.schedules[semester] is None:
                continue
            filename = self._schedule_filenames.get(semester, None)

            if filename is None or filename == "":
                filename = self.gui.select_file_to_save(f"Save Schedule As ({semester.name.upper()})")

            if filename is not None and filename != "":
                self.schedules[semester].write_file(filename)
                self.dirty_flag = False
                self.schedule_filename(semester,filename)

    # ============================================================================================
    # Event handlers - exit
    # ============================================================================================
    def exit_event(self, *_):
        """program is exiting"""
        if self.dirty_flag:
            ans = self.gui.ask_yes_no("File", "Save File?")
            if ans:
                self.save_schedule()

    def menu_exit_event(self, _:SemesterType):
        self.gui.exit_schedule()

    # ============================================================================================
    # Event handlers - go
    # ============================================================================================
    def go(self):

        for semester in VALID_SEMESTERS:
            if (self.gui.selected_files[semester].get() == "" or
                self.gui.selected_files[semester].get() == str(None)):
                self.new_menu_event(semester)
                self.set_dirty_method(False)
            else:
                self._open_file(self.gui.selected_files[semester].get(), semester)
        self.standard_page = self.gui.create_standard_page(self._notebook_tabs)
        self.set_dirty_indicator()

    # ==================================================================
    # notebook tab has changed
    # ==================================================================
    def notebook_tab_has_changed(self, name: str, frame):
        """
        the standard page notebook tab has been change
        :param name: name of the tab
        :param frame: container where the gui is stored
        """
        self.current_tab = name
        if name == "fall":
            self.gui.select_tab(f"fall {self.NB_allocation}")
            self._allocation_manager_already_open = False
        if name == "winter":
            self.gui.select_tab(f"winter {self.NB_allocation}")
            self._allocation_manager_already_open = False
        for semester in VALID_SEMESTERS:
            if name == f"{semester.name} {self.NB_allocation}":
                self.update_allocation(frame, semester)
            if name == f"{semester.name} {self.NB_course}":
                self.update_edit_courses(frame, semester)
                self._allocation_manager_already_open = False
            elif name == f"{semester.name} {self.NB_teacher}":
                self.update_edit_teachers(frame, semester)
                self._allocation_manager_already_open = False
            elif name == f"{semester.name} {self.NB_students}":
                self.update_edit_students(frame, semester)
                self._allocation_manager_already_open = False

    # ==================================================================
    # update the allocation frame
    # ==================================================================
    def update_allocation(self, frame, semester):
        if not self._allocation_manager_already_open:
            other_schedules = [self.schedules[s] for s in VALID_SEMESTERS if s != semester]
            AllocationEditor(
                self.set_dirty_method,
                frame,
                schedule=self.schedules[semester],
                other_schedules = other_schedules
            )
        self._allocation_manager_already_open = True

    # ==================================================================
    # draw_edit_courses
    # ==================================================================
    def update_edit_courses(self, frame, semester):
        """A page where courses can be added/modified or deleted"""
        data_entry = EditCourses(self.set_dirty_method, frame, self.schedules[semester])
        data_entry.schedule = self.schedules[semester]
        data_entry.refresh()

    # ==================================================================
    # update_edit_teachers
    # ==================================================================
    def update_edit_teachers(self, frame, semester):
        """A page where teacher can be added/modified or deleted"""
        data_entry = EditResources(self.set_dirty_method, frame, ResourceType.teacher,
                                   self.schedules[semester], self.preferences)
        data_entry.schedule = self.schedules[semester]
        data_entry.refresh()

    # ==================================================================
    # update_edit_students
    # ==================================================================
    def update_edit_students(self, frame, semester):
        data_entry = StudentNumbers(self.set_dirty_method, frame, self.schedules[semester])
        data_entry.refresh()

    # ==================================================================
    # schedule has been modified, update gui as required
    # ==================================================================
    def set_dirty_method(self, value: Optional[bool] = None) -> bool:

        # if value is true, and autosave is on, save the file
        if value and self.preferences.auto_save():
            self.save_schedule()
            value = False

        if value is not None:
            self.dirty_flag = value
        return self.dirty_flag

    # ============================================================================================
    # Event handler, auto save setting changed
    # ============================================================================================
    def set_dirty_indicator(self):
        if self.gui and self.dirty_flag:
            self.gui.dirty_text = "UNSAVED"
        else:
            self.gui.dirty_text = self.auto_save_text

    def auto_save_set(self, value, *_):
        self.preferences.auto_save(value)
        self.set_dirty_indicator()
        self.preferences.save()

    @property
    def auto_save_text(self):
        if self.preferences.auto_save():
            return "AUTO SAVE ON"
        else:
            return ""





import inspect
import os
import tkinter
from os import path

import pytest

from src.scheduling_and_allocation.Utilities import Preferences
from src.scheduling_and_allocation.gui_pages.scheduler_tk import SchedulerTk, MAIN_PAGE_EVENT_HANDLERS
from src.scheduling_and_allocation.presenter.menus_main_menu_scheduler import MAIN_MENU_EVENT_HANDLERS
from src.scheduling_and_allocation.presenter.scheduler import Scheduler

SCHEDULE_FILE = path.dirname(__file__) + "/data_test.csv"
PREVIOUS_FILE = path.dirname(__file__) + "/data_test_prev.csv"
BIN_DIR = path.dirname(__file__) + "/../../schedule"
BAD_SCHEDULE_FILE = path.dirname(__file__) + "/data_bad_test.csv"
CREATED_SCHEDULE_FILE = path.dirname(__file__) + "/data_new_test.csv"

# ============================================================================
# overload the gui
# ============================================================================
CURRENT_TEST_FILENAME = SCHEDULE_FILE


class SchedulerTkTest(SchedulerTk):
    def __init__(self, *args, **kwargs):
        self.called: dict[str, bool] = {}
        preferences: Preferences = Preferences()
        super().__init__("Testing", preferences, BIN_DIR)

    def clear(self):
        self.called.clear()

    def show_custom_message(self, title="", msg=""):
        self.called["show_error"] = True

    def start_event_loop(self):
        self.called["start_event_loop"] = True

    def show_error(self, title: str, msg: str, detail: str = ""):
        self.called["show_error"] = True

    def show_message(self, title: str, msg: str, detail: str = ""):
        self.called["show_message"] = True

    def select_file_to_open(self, title:str) -> str:
        self.called["select_file_to_open"] = True
        return CURRENT_TEST_FILENAME

    def select_file_to_save(self) -> str:
        self.called["select_file_to_save"] = True
        return CREATED_SCHEDULE_FILE

    def ask_yes_no(self, title: str, msg: str, detail: str = ""):
        return True

    def create_menu_and_toolbars(*args, **kwargs): ...

    def create_front_page(*args, **kwargs): ...

    def create_standard_page(*args, **kwargs): ...

    def create_status_bar(*args, **kwargs): ...

    def create_welcome_page(self, semester): ...


# ============================================================================
# tests
# ============================================================================
@pytest.fixture()
def gui():
    gui_test = SchedulerTkTest()
    gui_test.clear()
    return gui_test


def test_init(gui):
    obj = Scheduler(BIN_DIR, gui)
    assert gui.called["start_event_loop"]
    assert obj.schedule is None
    assert obj.schedule_filename is ''


# =============================================================================
# verify that menu event handlers have an 'event=None' as part of their signature
# =============================================================================
def test_menu_event_handlers_have_correct_signature(gui):
    Scheduler(BIN_DIR, gui)
    print("")
    for name, handler in MAIN_MENU_EVENT_HANDLERS.items():
        print(name)
        sig = inspect.signature(handler)
        print(sig)
        assert len(sig.parameters) == 0


# =============================================================================
# new, open
# =============================================================================
def test_new_from_menu(gui):
    """open new file from menu
    a) dirty flag set
    b) schedule filename set to ''
    c) gui schedule filename set  to ''
    d) gui previous filename remains unchanged
    """

    # prepare
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj.dirty_flag = False

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_new"]()

    # verify
    assert obj.dirty_flag
    assert obj.schedule is not None
    assert gui.schedule_filename == ""
    assert len(obj.schedule.teachers()) == 0


def test_new_button_press_event(gui):
    """same event handling for button press and menu"""

    # prepare
    Scheduler(BIN_DIR, gui)

    # verify
    assert MAIN_PAGE_EVENT_HANDLERS["file_new"] == MAIN_MENU_EVENT_HANDLERS["file_new"]


def test_file_open_from_menu(gui):
    """open file CURRENT_TEST_FILENAME from menu
    a) dirty flag is unset
    b) schedule filename set to file name that was opened
    b) gui schedule filename set to file that was opened
    c) schedule was read
    """

    # prepare
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj.dirty_flag = True

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_open"]()

    # verify
    assert not obj.dirty_flag
    assert gui.called.get("select_file_to_open", None)
    assert gui.schedule_filename == CURRENT_TEST_FILENAME
    assert not gui.called.get("show_error", None)
    assert obj.schedule_filename == CURRENT_TEST_FILENAME
    assert obj.schedule is not None
    assert len(obj.schedule.teachers()) != 0


def test_bad_file_open_from_menu(gui):
    """open non-existent file from menu (after a previous file was opened)
    a) dirty flag unchanged (not unset)
    b) schedule filename is unchanged
    c) schedule object is unchanged
    """

    # prepare
    global CURRENT_TEST_FILENAME
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_open"]()
    obj.dirty_flag = True
    old_schedule = obj.schedule
    old_filename = obj.schedule_filename
    old_dirty_flag = obj.dirty_flag
    CURRENT_TEST_FILENAME = BAD_SCHEDULE_FILE
    MAIN_MENU_EVENT_HANDLERS["file_open"]()

    # verify
    assert obj.dirty_flag == old_dirty_flag
    assert gui.called.get("select_file_to_open", None)
    assert gui.schedule_filename == old_filename
    assert gui.called.get("show_error", None) is True
    assert obj.schedule_filename == old_filename
    assert obj.schedule is old_schedule


def test_bad_file_open_from_menu2(gui):
    """open non-existent file from menu (no previous file opened)
    a) dirty flag unchanged (not unset)
    b) schedule filename is unchanged
    c) schedule object is unchanged
    """

    # prepare
    global CURRENT_TEST_FILENAME
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj.dirty_flag = True

    # execute
    old_schedule = obj.schedule
    old_filename = obj.schedule_filename
    old_dirty_flag = obj.dirty_flag
    old_gui_filename = obj.gui.schedule_filename
    CURRENT_TEST_FILENAME = BAD_SCHEDULE_FILE
    MAIN_MENU_EVENT_HANDLERS["file_open"]()

    # verify
    assert obj.dirty_flag == old_dirty_flag
    assert gui.called.get("select_file_to_open", None)
    assert gui.schedule_filename == old_gui_filename
    assert gui.called.get("show_error", None) is True
    assert obj.schedule_filename == old_filename
    assert obj.schedule is old_schedule


def test_open_from_menu_user_cancels(gui):
    """start to open a schedule, but then user cancels (no previous file open)
    a) dirty flag unchanged (not unset)
    b) schedule filename is unchanged
    c) schedule object is unchanged
    """

    # prepare
    global CURRENT_TEST_FILENAME
    CURRENT_TEST_FILENAME = SCHEDULE_FILE
    obj = Scheduler(BIN_DIR, gui)
    obj.dirty_flag = True
    old_dirty_flag = obj.dirty_flag
    old_schedule = obj.schedule
    old_filename = obj.schedule_filename
    old_gui_filename = obj.gui.schedule_filename

    # execute
    CURRENT_TEST_FILENAME = None
    MAIN_MENU_EVENT_HANDLERS["file_open"]()

    # verify
    assert obj.dirty_flag == old_dirty_flag  # didn't open new file, dirty flag unchanged
    assert gui.called.get("select_file_to_open", None)
    assert gui.schedule_filename == old_gui_filename
    assert not gui.called.get("show_error", None)
    assert obj.schedule_filename == old_filename  # didn't open new file, old file still valid
    assert obj.schedule == old_schedule  # didn't open new file, schedule is the old one


def test_open_from_menu_user_cancels2(gui):
    """start to open a schedule, but then user cancels (previous file open)
    a) dirty flag unchanged
    b) schedule filename is unchanged
    c) schedule object is unchanged
    """

    # prepare
    global CURRENT_TEST_FILENAME
    CURRENT_TEST_FILENAME = SCHEDULE_FILE
    obj = Scheduler(BIN_DIR, gui)
    MAIN_MENU_EVENT_HANDLERS["file_open"]()
    obj.dirty_flag = True
    old_dirty_flag = obj.dirty_flag
    old_schedule = obj.schedule
    old_filename = obj.schedule_filename
    old_gui_filename = obj.gui.schedule_filename

    # execute
    CURRENT_TEST_FILENAME = None
    MAIN_MENU_EVENT_HANDLERS["file_open"]()

    # verify
    assert not gui.called.get("show_error", None)
    assert gui.called.get("select_file_to_open", None)

    assert obj.dirty_flag == old_dirty_flag  # didn't open new file, dirty flag unchanged
    assert gui.schedule_filename == old_filename
    assert obj.gui.schedule_filename == old_gui_filename
    assert obj.schedule == old_schedule  # didn't open new file, schedule is the old one


def test_open_from_main_page_event(gui):
    """same event handling for button press and menu"""

    # prepare
    Scheduler(BIN_DIR, gui)

    # verify
    assert MAIN_PAGE_EVENT_HANDLERS["file_open"] == MAIN_MENU_EVENT_HANDLERS["file_open"]


def test_open_previous(gui):
    """Using the preferences, get previous file that was opened, and open that
    - filename set to PREVIOUS FILE
    - dirty flag unset
    - schedule has 18 teachers
    """

    # prepare
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj.dirty_flag = True

    # execute
    print(PREVIOUS_FILE)
    obj.preferences.previous_file(PREVIOUS_FILE)
    obj.preferences.save()
    MAIN_PAGE_EVENT_HANDLERS["file_open_previous"]()

    # verify
    assert not obj.dirty_flag
    assert gui.schedule_filename == PREVIOUS_FILE
    assert not gui.called.get("show_error", None)
    assert obj.schedule_filename == PREVIOUS_FILE
    assert obj.schedule is not None
    assert len(obj.schedule.teachers()) == 18

    # prepare
    global CURRENT_TEST_FILENAME
    CURRENT_TEST_FILENAME = SCHEDULE_FILE
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj.dirty_flag = False

    # execute
    MAIN_PAGE_EVENT_HANDLERS["file_open"]()

    # verify
    assert not obj.dirty_flag
    assert gui.schedule_filename == SCHEDULE_FILE
    assert not gui.called.get("show_error", None)
    assert obj.schedule_filename == SCHEDULE_FILE
    assert obj.schedule is not None
    assert len(obj.schedule.teachers()) != 0


# =============================================================================
# save, save as
# =============================================================================
def test_save_from_menu(gui):
    """open file, and save
    - dirty flag is unset
    - new file is created
    """

    # prepare
    global CURRENT_TEST_FILENAME
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj._open_file(SCHEDULE_FILE)
    assert obj.schedule is not None
    before = len(obj.schedule.teachers())
    last_modified_time = os.path.getmtime(SCHEDULE_FILE)
    obj.dirty_flag = True

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_save"]()

    # verify
    assert not obj.dirty_flag
    assert os.path.getmtime(SCHEDULE_FILE) > last_modified_time
    assert obj.schedule_filename == SCHEDULE_FILE
    obj._open_file(SCHEDULE_FILE)
    assert len(obj.schedule.teachers()) == before


def test_save_as_from_menu(gui):
    """open file, save as
    - dirty flag is unset
    - ask user for filename
    - filename is changed
    """

    # prepare
    global CURRENT_TEST_FILENAME
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    obj._open_file(SCHEDULE_FILE)
    assert obj.schedule is not None
    before = len(obj.schedule.teachers())
    obj.dirty_flag = True
    try:
        os.remove(CREATED_SCHEDULE_FILE)
    except FileNotFoundError:
        pass

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_save_as"]()

    # verify
    assert not obj.dirty_flag
    assert gui.called.get("select_file_to_save", None)
    assert os.path.exists(CREATED_SCHEDULE_FILE)
    assert obj.schedule_filename == CREATED_SCHEDULE_FILE
    assert obj.gui.schedule_filename == CREATED_SCHEDULE_FILE
    obj._open_file(CREATED_SCHEDULE_FILE)
    assert len(obj.schedule.teachers()) == before


def test_save_no_schedule(gui):
    """no schedule, save
    - error message is displayed
    - dirty change remains same as before
    """
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    last_modified_time = os.path.getmtime(SCHEDULE_FILE)
    assert obj.schedule_filename == ""
    obj.dirty_flag = True

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_save"]()

    # verify
    assert obj.dirty_flag  # no schedule to save, no change to dirty flag
    assert os.path.getmtime(SCHEDULE_FILE) == last_modified_time
    assert obj.schedule_filename == ''
    assert gui.called.get("show_error")


def test_save_as_no_schedule(gui):
    """no schedule, save as
    - error message is displayed
    - dirty change remains same as before
    """
    obj = Scheduler(BIN_DIR, gui)
    assert obj.schedule is None
    last_modified_time = os.path.getmtime(SCHEDULE_FILE)
    assert obj.schedule_filename is ''
    obj.dirty_flag = True

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_save"]()

    # verify
    assert obj.dirty_flag  # no file to save, so dirty flag not updated
    assert os.path.getmtime(SCHEDULE_FILE) == last_modified_time
    assert obj.schedule_filename is ''
    assert gui.called.get("show_error", None)
    assert not gui.called.get("select_file_to_save", None)


def test_save_no_filename(gui):
    """new schedule, save
    - asks user for filename
    - dirty flag unset
    """
    # prepare
    obj = Scheduler(BIN_DIR, gui)
    MAIN_MENU_EVENT_HANDLERS["file_new"]()
    assert obj.schedule_filename == ''
    obj.dirty_flag = True
    try:
        os.remove(CREATED_SCHEDULE_FILE)
    except FileNotFoundError:
        pass

    # execute
    MAIN_MENU_EVENT_HANDLERS["file_save"]()

    # verify
    assert not obj.dirty_flag
    assert os.path.exists(CREATED_SCHEDULE_FILE)
    assert obj.schedule_filename == CREATED_SCHEDULE_FILE
    assert not gui.called.get("show_error", None)
    assert gui.called.get("select_file_to_save", None)


# =============================================================================
# exit
# =============================================================================
def test_exit_with_new_schedule(gui):
    """new schedule, exit
    - asks user save?
    - asks user for filename
    """
    # prepare
    obj = Scheduler(BIN_DIR, gui)
    MAIN_MENU_EVENT_HANDLERS["file_new"]()
    assert obj.schedule_filename == ''
    obj.dirty_flag = True
    try:
        os.remove(CREATED_SCHEDULE_FILE)
    except FileNotFoundError:
        pass

    # execute
    MAIN_PAGE_EVENT_HANDLERS["file_exit"]()

    # verify
    assert os.path.exists(CREATED_SCHEDULE_FILE)
    assert obj.schedule_filename == CREATED_SCHEDULE_FILE
    assert not gui.called.get("show_error", None)
    assert gui.called.get("select_file_to_save", None)


# =============================================================================
# changing semesters
# =============================================================================
def test_semester_switch(gui):
    """switch semesters
    - previous filename is modified
    """

    # prepare
    obj = Scheduler(BIN_DIR, gui)
    obj.preferences.semester("fall")
    obj.preferences.previous_file("data_fall.csv")
    obj.preferences.save()
    obj.preferences.semester("winter")
    obj.preferences.previous_file("data_winter.csv")
    obj.preferences.save()

    # execute
    obj.gui.current_semester = "fall"
    MAIN_PAGE_EVENT_HANDLERS["semester_change"]()

    # validate
    assert os.path.basename(obj.previous_filename) == "data_fall.csv"

    # execute
    obj.gui.current_semester = "winter"
    MAIN_PAGE_EVENT_HANDLERS["semester_change"]()

    # validate
    assert os.path.basename(obj.previous_filename) == "data_winter.csv"



from typing import Callable, Optional

import pytest
from _pytest.fixtures import fixture

from src.scheduling_and_allocation.gui_generics.menu_and_toolbars import MenuItem, MenuType
from src.scheduling_and_allocation.model import ResourceType, ConflictType, Block, Schedule, WeekDay, SemesterType
from src.scheduling_and_allocation.presenter.view import View


def _default_menu(*_) -> list[MenuItem]:
    menu = MenuItem(name="nothing", label="nothing", menu_type=MenuType.Command, command=lambda: None)
    return [menu, ]

# =====================================================================================================================
# View Gui Test class
# =====================================================================================================================

class ViewDynamicTkTest:

    def __init__(self, frame, title: str,resource_type: ResourceType):
        # ------------------------------------------------------------------------------------------------------------
        # handlers
        # ------------------------------------------------------------------------------------------------------------
        # returns menu info for gui block popup menu
        self.get_popup_menu_handler: Callable[[str], list[MenuItem]] = _default_menu

        # redraws all the appropriate blocks on the canvas
        self.refresh_blocks_handler: Callable[[], None] = lambda: None

        # handles the closing of this view
        self.on_closing_handler: Callable = lambda *_: None

        # handles a double click on a gui block
        self.double_click_block_handler: Callable[[str], None] = lambda gui_id: None

        # handles the movement of a gui block
        self.gui_block_is_moving_handler: Callable[[str, float, float], None] = lambda gui_id, day, start: None

        # handles the dropping of a gui block
        self.gui_block_has_dropped_handler: Callable[[str], None] = lambda gui_id: None


        # undo/redo handlers
        self.undo_handler: Callable[[], None] = lambda: None
        self.redo_handler: Callable[[], None] = lambda: None

        self.draw_blocks_info = {}
        self.colour_blocks_info = {}

    def get_scale(self):
        return 1

    def draw(self, scale_factor: float = 1.0):
        self.refresh_blocks_handler()


    def draw_block(self, resource_type: ResourceType, day: int, start_time: float,
                   duration: float, text: str, gui_block_id, movable=True):
        self.draw_blocks_info[gui_block_id] = (day, start_time, duration, movable)

    def colour_block(self, gui_block_id: str, resource_type: ResourceType, is_movable=True,
                     conflict: ConflictType = None):
        self.colour_blocks_info[gui_block_id] = (resource_type, is_movable, conflict)

    def move_gui_block(self, gui_block_id: str, day_number, start_number):
        pass

    def modify_movable(self, gui_id, flag):
        pass


# =====================================================================================================================
# Views Controller
# =====================================================================================================================
class ViewsControllerTest:

    def __init__(self, dirty_flag_method, frame, schedule):
        self.notified_toggled = None
        self.remove_redoes = False
        self.action_toggle = None
        self.on_closing_called = False
        self.notify_move_block_to_resource_args = {}
        self.open_companion_view_called = None
        self.action_resource_saved = {}
        self.block_move = {}
        self.action_moved = {}
        self.undo_called = False
        self.redo_called = False

    def notify_block_move(self, resource_number: Optional[str], moved_block: Block, day: float, start_time:float):
        self.block_move["id"] = resource_number
        self.block_move["block"] = moved_block
        self.block_move["day"] = day
        self.block_move["start"] = start_time


    def notify_block_movable_toggled(self, modified_block):
        self.notified_toggled = modified_block

    def notify_move_block_to_resource(self, resource_type: ResourceType, block, from_resource, to_resource):
        self.notify_move_block_to_resource_args["block"] = block
        self.notify_move_block_to_resource_args["from"] = from_resource
        self.notify_move_block_to_resource_args["to_resource"] = to_resource
        self.notify_move_block_to_resource_args["type"] = resource_type

    def open_companion_view(self, block):
        self.open_companion_view_called = block

    def view_is_closing(self, resource):
        self.on_closing_called = True

    def save_action_block_move(self, block, from_day:float, to_day:float, from_time:float, to_time:float):
        self.action_moved['block'] = block
        self.action_moved['from'] = from_day
        self.action_moved['to'] = to_day
        self.action_moved['from_time'] = from_time
        self.action_moved['to_time'] = to_time

    def save_action_block_resource_changed(self, resource_type: ResourceType, block, from_resource, to_resource):
        self.action_resource_saved["block"] = block
        self.action_resource_saved["from"] = from_resource
        self.action_resource_saved["to_resource"] = to_resource
        self.action_resource_saved["type"] = resource_type

    def save_action_block_movable_toggled(self, block, was: bool):
        self.action_toggle = (block, was)

    def remove_all_redoes(self):
        self.remove_redoes = True

    def undo(self):
        self.undo_called = True

    def redo(self):
        self.redo_called = True


# =====================================================================================================================
# fixtures
# =====================================================================================================================
@pytest.fixture()
def schedule_obj():
    schedule = Schedule()

    # teachers
    t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
    t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
    t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
    t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id = "4")

    # labs
    l1 = schedule.add_update_lab("P107", "C-Lab")
    l2 = schedule.add_update_lab("P322", "Mac Lab")
    l3 = schedule.add_update_lab("P325")
    l4 = schedule.add_update_lab("BH311", "Britain Hall")

    # streams
    st1 = schedule.add_update_stream("1A", "calculus available")
    st2 = schedule.add_update_stream("1B", "no calculus")
    st3 = schedule.add_update_stream("2A")
    st4 = schedule.add_update_stream("2B")

    # courses
    c_001 = schedule.add_update_course("001","BasketWeaving", SemesterType.fall )
    s_001_1 = c_001.add_section("1", section_id=1)
    s_001_1.add_stream(st1)
    b_001_1_1 = s_001_1.add_block(WeekDay.Monday,  8 )
    b_001_1_2 = s_001_1.add_block(WeekDay.Monday,  10 )
    b_001_1_1.add_teacher(t1)
    b_001_1_2.add_teacher(t1)
    b_001_1_1.add_lab(l1)
    b_001_1_2.add_lab(l1)

    s_001_2 = c_001.add_section("2",section_id=2)
    s_001_2.add_stream(st2)
    b_001_2_1 = s_001_2.add_block(WeekDay.Tuesday,  8 )
    b_001_2_2 = s_001_2.add_block(WeekDay.Tuesday,  10 )
    b_001_2_1.add_teacher(t2)
    b_001_2_2.add_teacher(t3)
    b_001_2_1.add_lab(l2)
    b_001_2_2.add_lab(l3)


    c_002 = schedule.add_update_course("002","Thumb Twiddling", SemesterType.fall )
    s_002_1 = c_002.add_section("1",)
    b_002_1_1 = s_002_1.add_block(WeekDay.Monday,  8 )
    b_002_1_2 = s_002_1.add_block(WeekDay.Monday,  10 )
    s_002_2 = c_002.add_section("2",)
    b_002_2_1 = s_002_2.add_block(WeekDay.Tuesday,  8 )
    b_002_2_2 = s_002_2.add_block(WeekDay.Tuesday,  10 )



    return schedule

DIRTY = False
def dirty_flag_method(flag = None):
    global DIRTY
    if flag is not None:
        DIRTY = flag
    return DIRTY

@fixture
def gui():
    gui = ViewDynamicTkTest("frame", "title", ResourceType.none)
    return gui

@fixture
def view_control():
    vc = ViewsControllerTest("", "", "")
    return vc

# ===================================================================================================================
# Testing View
# ===================================================================================================================
def test_init(schedule_obj, view_control, gui):
    """Create a view    """

    # execute
    #     def __init__(self, views_controller: ViewsController, frame, schedule: Schedule,
    #                  resource: RESOURCE, dirty_flag_method:Callable, gui:ViewDynamicTk=None,
    #                  ):
    view = View(view_control,"", schedule_obj, resource=schedule_obj.get_teacher_by_name("Jane", "Doe"),
                dirty_flag_method = dirty_flag_method, gui=gui )


def test_draw_blocks_called_during_init(schedule_obj, view_control, gui):
    """Create a view
    1. all blocks have a gui_id
    2. each gui block is drawn with the appropriate parameters
    3. each gui block is told to colour with each of the appropriate parameters
    """
    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    blocks[0].add_conflict(ConflictType.LUNCH)


    # execute: (view is drawn after block has been modified)
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )

    # verify
    for block in blocks:
        assert block in view.gui_blocks.values()

    for gui_id in view.gui_blocks:
        block: Block = view.gui_blocks[gui_id]
        assert gui_id in gui.draw_blocks_info
        assert gui.draw_blocks_info[gui_id] == (block.day.value, block.start,
                                                block.duration, block.movable)
        assert gui.colour_blocks_info[gui_id] == (ResourceType.teacher, block.movable, block.conflict )


def test_draw_blocks(schedule_obj, view_control, gui):
    """draw all the blocks
    1. all blocks have a gui_id
    2. each gui block is drawn with the appropriate parameters
    3. each gui block is told to colour with each of the appropriate paramters
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    blocks[0].add_conflict(ConflictType.LUNCH)

    # execute
    view.draw_blocks()

    # verify
    for block in blocks:
        assert block in view.gui_blocks.values()

    for gui_id in view.gui_blocks:
        block: Block = view.gui_blocks[gui_id]
        assert gui_id in gui.draw_blocks_info
        assert gui.draw_blocks_info[gui_id] == (block.day.value, block.start,
                                                block.duration, block.movable)
        assert gui.colour_blocks_info[gui_id] == (ResourceType.teacher, block.movable, block.conflict )

def test_view_keeps_track_of_blocks_and_gui_ids(schedule_obj, view_control, gui):
    """Create the view
    1. Able to determine if a certain block is in the view
    """
    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    block = schedule_obj.get_blocks_for_teacher(teacher)[0]

    # execute
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )

    # validate
    assert view.is_block_in_view(block)

# ===================================================================================================================
# Testing View Popup and Popup Handlers
# ===================================================================================================================
def test_popup_menu(schedule_obj, view_control, gui):
    """Invoke pop-up menu
    1. returns a MenuItem list
    2. has a set/unset 'movable' option on list
    3. has move resource to... on list
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]

    # execute
    menu = gui.get_popup_menu_handler(gui_id)

    # verify
    assert len(menu) != 0
    mi = None
    for m in menu:
        if "movable" in m.label:
            mi = m.command
    assert mi is not None

    mi: Optional[MenuItem] = None
    for m in menu:
        if "Move class to" in m.label:
            mi = m
    assert mi is not None

    assert "Bugs Bunny" in (m.label for m in mi.children)

def test_toggle_movable(schedule_obj, view_control, gui):
    """from pop-up menu, toggle block movability
    1. modify the block accordingly
    2. notify view controller that block has changed
    3. remove all 'redoes' because user had done something
    4. add action to the 'undo' list
    5. dirty flag is set
    """

    # prepare
    dirty_flag_method(False)
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]
    block.movable = True
    menu = gui.get_popup_menu_handler(gui_id)

    # execute
    for m in menu:
        if "movable" in m.label:
            m.command()

    # verify
    assert not block.movable
    assert dirty_flag_method()

    # execute
    view.toggle_is_movable(block, gui_id)

    # verify
    assert block.movable
    assert view_control.remove_redoes
    assert view_control.notified_toggled == block
    assert view_control.action_toggle == (block, False)

def test_move_block_to_different_resource(schedule_obj, view_control, gui):
    """from pop-up menu, move block to different resource
    1. notify view controller that block has changed
    2. remove all 'redoes' because user had done something
    3. add action to the 'undo' list
       NOTE: block is not changed in view, it is changed in the view_controller
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]
    block.movable = True
    menu = gui.get_popup_menu_handler(gui_id)

    # execute
    mi: Optional[MenuItem] = None
    for m in menu:
        if "Move class to" in m.label:
            mi = m
    assert mi is not None

    for m in mi.children:
        if "John Doe" in m.label:
            m.command()

    # verify
    assert len(view_control.notify_move_block_to_resource_args) != 0
    jd = schedule_obj.get_teacher_by_name("John","Doe")
    assert view_control.remove_redoes

    assert view_control.notify_move_block_to_resource_args["block"] == block
    assert view_control.notify_move_block_to_resource_args["from"] == teacher
    assert view_control.notify_move_block_to_resource_args["to_resource"] == jd
    assert view_control.notify_move_block_to_resource_args["type"] == ResourceType.teacher

    assert view_control.action_resource_saved["block"] == block
    assert view_control.action_resource_saved["from"] == teacher
    assert view_control.action_resource_saved["to_resource"] == jd
    assert view_control.action_resource_saved["type"] == ResourceType.teacher



def test_on_closing_informs_view_controller(schedule_obj, view_control, gui):
    """close the view
    1. view controller is informed
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )

    # execute
    view.on_closing()

    # verify
    assert view_control.on_closing_called

def test_handle_double_click(schedule_obj, view_control, gui):
    """Block is doubled clicked on view
    1. View control is asked to pop up companion views
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]

    # execute
    gui.double_click_block_handler(gui_id)

    # verify
    assert view_control.open_companion_view_called == block

def test_gui_block_is_moving(schedule_obj, view_control, gui):
    """block is moving
    1. view_control is informed
    2. block is moving (snapping to time as it goes)
    3. conflicts are recalculated
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]
    o_day = block.day.value

    # execute
    gui.gui_block_is_moving_handler( gui_id,  block.day.value-0.2, 9.45)

    # validate
    assert view_control.block_move['block'] == block
    assert view_control.block_move['id'] == teacher.number
    assert view_control.block_move['day'] == block.day.value-0.2

    assert block.day.value == o_day
    assert block.start == 9.50

    assert block.conflict


def test_gui_block_has_dropped(schedule_obj, view_control, gui):
    """block has dropped
    1. view_control is informed
    2. block is moving (snapping to time as it goes)
    3. conflicts are recalculated
    4. action has been saved
    5. redo has cleared
    6. gui block has been updated (don't know how to test this)
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]
    o_day = block.day
    o_start = block.start
    gui.gui_block_is_moving_handler( gui_id,  3, 9.45)

    # execute
    gui.gui_block_has_dropped_handler( gui_id)

    # validate
    assert block.day.value == 3
    assert block.start == 9.5

    assert block.conflict

    assert view_control.remove_redoes
    assert view_control.action_moved['block'] == block
    assert view_control.action_moved['from'] == o_day.value
    assert view_control.action_moved['to'] == 3
    assert view_control.action_moved['from_time'] == o_start
    assert view_control.action_moved['to_time'] == 9.5

def test_gui_block_has_dropped_but_not_moved(schedule_obj, view_control, gui):
    """block has dropped, but wasn't moved
    1. view_control is not called
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )
    gui_id = list(view.gui_blocks.keys())[0]
    block = view.gui_blocks[gui_id]

    # execute
    gui.gui_block_has_dropped_handler( gui_id)

    # validate
    assert not view_control.remove_redoes
    assert len(view_control.action_moved) == 0

def test_gui_undo_passed_to_controller(schedule_obj, view_control, gui):
    """user triggered a undo
    1. undo trigger is passed to controller
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )

    # execute
    gui.undo_handler()

    # validate
    assert view_control.undo_called

def test_gui_redo_passed_to_controller(schedule_obj, view_control, gui):
    """user triggered a redo
    1. redo trigger is passed to controller
    """

    # prepare
    teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    view = View(view_control,"", schedule_obj, resource=teacher,
                dirty_flag_method = dirty_flag_method, gui=gui )

    # execute
    gui.redo_handler()

    # validate
    assert view_control.redo_called




import re
from dataclasses import dataclass

from ..gui_pages.allocation_grid_tk import AllocationGridTk
from ..model import Schedule, Course, Section, Teacher
from ..ci_calculator.ci_calculation import calculate_ci

# =====================================================================================================================
# InnerData and SummaryRow data classes
# =====================================================================================================================
@dataclass
class InnerData:
    teacher: Teacher
    course: Course
    section: Section
    hours: float

@dataclass
class SummaryRow:
    release: str
    total_hrs: str
    semester_ci: str
    year_ci: str
    teacher:Teacher

# =====================================================================================================================
# Allocation Editor
# =====================================================================================================================
class AllocationEditor:

    # -----------------------------------------------------------------------------------------------------------------
    # constructor
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, set_dirty_flag, frame, schedule: Schedule, other_schedules: list[Schedule] = None):
        """
        Add teachers to course/sections, specifying hours.
        NOTE: Teachers will be added to all blocks if there are blocks,
        :param set_dirty_flag: method to set dirty flag
        :param frame: container where to draw gui stuff
        :param schedule: schedule
        :param other_schedules: schedules that are not part of this semester (used to calculate total CI)
        """
        self.set_dirty_flag = set_dirty_flag
        self.frame = frame
        self.schedule = schedule
        self.other_schedules = [] if other_schedules is None else other_schedules

        self.teachers = schedule.teachers()
        self.courses = schedule.courses_with_allocation()


        self.remaining_text = "Avail Hrs"
        self.summary_headings = ["RT", "hrs", "CI", "YEAR"]
        self.inner_data: dict[tuple[int,int],InnerData] = {}
        self.data_numbers_only: dict[tuple[int,int], float] = {}

        self.gui = AllocationGridTk(frame,
                        rows=len(self.teachers),
                        col_merge=[c.number_of_sections() for c in self.courses] ,
                        summary_merge = [len(self.summary_headings)],
                        cb_process_data_change=lambda *args: self.data_change_handler(*args),
                        bottom_cell_valid = lambda c: float(c) == 0.0,
                        )
        self.populate()


    # -----------------------------------------------------------------------------------------------------------------
    # using schedule, construct data that can be used to populate the gui allocation grid
    # -----------------------------------------------------------------------------------------------------------------
    def populate(self):

        teachers = self.teachers
        courses = self.courses

        teachers_text = list(map(lambda a: f"{a.firstname} {a.lastname[0:1]}.", teachers))
        courses_text = list(map(lambda a: str(re.sub(r'\s*\d\d\d-', '', a.number)), courses))
        courses_balloon = list(map(lambda a: f" {a.name} ({a.hours_per_week})" , courses))

        teacher_summaries:list[list[str]] = []
        data_numbers_only: dict[tuple[int,int], float] = {}

        # loop over the courses/sections, and save info into appropriate data structures
        col = 0
        sections_text: list[str] = []
        for course in courses:
            for section in course.sections():
                sections_text.append(section.number)

                # save info for each teacher ( for each course/sections)
                for row, teacher in enumerate(teachers):
                    self.inner_data[(row,col)]=InnerData(teacher=teacher,
                                                    course=course,
                                                    section=section,
                                                    hours=section.get_teacher_allocation(teacher))
                    data_numbers_only[(row,col)] = self.inner_data[row,col].hours

                col += 1

        # get the summary info for each teacher
        for row, teacher in enumerate(teachers):
            teacher_stats = self._calculate_summary(row)
            teacher_summaries.append([teacher_stats.release,
                                      teacher_stats.total_hrs,
                                      teacher_stats.semester_ci,
                                      teacher_stats.year_ci])

        # get unallocated hours for each course/section
        remaining_hours = AllocationEditor._calculate_unallocated_hours(self.inner_data)

        # add all the data to the gui Allocation Grid
        self.gui.populate(
            courses_text, courses_balloon, sections_text,
            teachers_text, data_numbers_only, [""], self.summary_headings,
            teacher_summaries, self.remaining_text,
            [f"{v:.1f}" for v in remaining_hours]
        )


    # -----------------------------------------------------------------------------------------------------------------
    # data change handler
    # -----------------------------------------------------------------------------------------------------------------
    def data_change_handler(self, row, col, value):

        # update schedule
        info = self.inner_data[row,col]
        self.inner_data[row,col].hours = float(value)
        info.section.set_teacher_allocation(self.teachers[row], float(value))

        summary = self._calculate_summary(row)
        self.gui.update_data('summary',row,0, summary.release)
        self.gui.update_data('summary',row,1, summary.total_hrs)
        self.gui.update_data('summary',row,2, summary.semester_ci)
        self.gui.update_data('summary',row,3, summary.year_ci)

        remaining_hours = AllocationEditor._calculate_unallocated_hours(self.inner_data)
        self.gui.update_data('bottom', 0, col, remaining_hours[col])
        self.set_dirty_flag(True)

    # -----------------------------------------------------------------------------------------------------------------
    # calculate the summary for a particular row (ci/ total hrs/ etc)
    # -----------------------------------------------------------------------------------------------------------------
    def _calculate_summary(self, row):
        teacher = self.teachers[row]
        hrs = 0
        for course in self.schedule.get_courses_for_teacher(teacher):
            for section in course.sections():
                hrs += section.get_teacher_allocation(teacher)
        semester_ci = calculate_ci(teacher=teacher, schedule=self.schedule)

        yearly_ci = semester_ci
        for other in self.other_schedules:
            other_teacher = other.get_teacher_by_name(teacher.firstname, teacher.lastname)
            if other_teacher is not None:
                yearly_ci += calculate_ci(other_teacher, schedule=other)

        # convert all the numbers into their appropriate string variations
        return (SummaryRow(release="" if teacher.release == 0 else f"{teacher.release:6.3f}",
                                            semester_ci="" if semester_ci == 0 else f"{semester_ci:6.1f}",
                                            teacher=teacher,
                                            total_hrs="" if hrs == 0 else f"{hrs:6.1f}",
                                            year_ci="" if yearly_ci == 0 else f"{yearly_ci:6.1f}"))


    # -----------------------------------------------------------------------------------------------------------------
    # calculate all the unallocated hours per section
    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _calculate_unallocated_hours(data:dict[tuple[int,int],InnerData]):
        total_per_col:dict[int, float] = {}
        req_per_col:dict[int,float] = {}
        remaining_hours:dict[int, float] = {}
        for location,datum in data.items():
            row,col = location
            total_per_col[col] = total_per_col.get(col, 0) + datum.hours
            req_per_col[col] = datum.course.hours_per_week
        for col in total_per_col.keys():
            remaining_hours[col] = req_per_col[col] - total_per_col[col]

        remaining = sorted(remaining_hours.items())
        return [value for key, value in remaining]





DEFAULT_X_OFFSET: float = 1
DEFAULT_Y_OFFSET: float = 1
DEFAULT_X_ORIGIN: float = 0
DEFAULT_Y_ORIGIN: float = 0
DEFAULT_X_WIDTH_SCALE: float = 100
DEFAULT_Y_HEIGHT_SCALE: float = 60
CURRENT_SCALE: float = 1.0

# =============================================================================================================
# Get a drawing scale (for schedule views) based off a factor
# =============================================================================================================
class DrawingScale:
    def __init__(self, factor=CURRENT_SCALE):
        self.x_offset: float = DEFAULT_X_OFFSET
        self.y_offset: float = DEFAULT_Y_OFFSET
        self.x_origin: float = DEFAULT_X_ORIGIN
        self.y_origin: float = DEFAULT_Y_ORIGIN
        self.x_width_scale: float = DEFAULT_X_WIDTH_SCALE * factor
        self.y_height_scale: float = DEFAULT_Y_HEIGHT_SCALE * factor
        self.current_scale: float = factor

